"""
Classes to work with account key material.
"""

from __future__ import annotations
import base58  # type: ignore
import base64
import hashlib
import unicodedata
from typing import Optional, Any
from Crypto.Cipher import AES  # type: ignore
from jsonschema import validate  # type: ignore
from neo3 import settings, vm
from neo3.contracts import abi, utils as contractutils, contract
from neo3.network.payloads import transaction, verification
from neo3.core import types, utils as coreutils, cryptography
from neo3.wallet import utils, scrypt_parameters as scrypt
from neo3.wallet.types import NeoAddress

# both constants below are used to encrypt/decrypt a private key to/from a nep2 key
NEP_HEADER = bytes([0x01, 0x42])
NEP_FLAG = bytes([0xE0])
# both constants are used when trying to decrypt a private key from a wif
WIF_PREFIX = bytes([0x80])
WIF_SUFFIX = bytes([0x01])
PRIVATE_KEY_LENGTH = 32


class MultiSigContext:
    """
    Signing context for use with multi signature accounts.
    """

    def __init__(self):
        #: indicates if the context has been initialised for usage.
        self.initialised = False
        #: minimum of signatures required for signing.
        self.signing_threshold = 999
        #: list of valid public keys for signing.
        self.expected_public_keys: list[cryptography.ECPoint] = []
        #: completed pairs.
        self.signature_pairs: dict[cryptography.ECPoint, bytes] = {}

    @property
    def is_complete(self):
        return len(self.signature_pairs) >= self.signing_threshold

    def signing_status(self) -> dict[cryptography.ECPoint, bool]:
        # shows which keys have been completed
        raise NotImplementedError

    def process_contract(self, script: bytes) -> None:
        valid, threshold, public_keys = contractutils.parse_as_multisig_contract(script)
        if not valid:
            raise ValueError("Invalid script")
        self.expected_public_keys = public_keys
        self.signing_threshold = threshold
        self.initialised = True


class AccountContract(contract.Contract):
    _contract_params_schema = {
        "type": ["object", "null"],
        "properties": {"name": {"type": "string"}, "type": {"type": "string"}},
        "required": ["name", "type"],
    }
    _json_schema = {
        "type": ["object", "null"],
        "properties": {
            "script": {"type": "string"},
            "parameters": {
                "type": "array",
                "items": _contract_params_schema,
                "minItems": 0,
            },
            "deployed": {"type": "boolean"},
        },
        "required": ["script", "parameters", "deployed"],
    }

    def __init__(
        self, script: bytes, parameter_list: list[abi.ContractParameterDefinition]
    ):
        super().__init__(script, [param.type for param in parameter_list])

        self.parameter_names: list[str] = [param.name for param in parameter_list]
        self.deployed: bool = False

    @classmethod
    def from_contract(cls, c: contract.Contract) -> AccountContract:
        if isinstance(c, AccountContract):
            return c

        parameters = [
            abi.ContractParameterDefinition(f"arg{index}", c.parameter_list[index])
            for index in range(len(c.parameter_list))
        ]
        return cls(script=c.script, parameter_list=parameters)

    @classmethod
    def from_json(cls, json: dict) -> AccountContract:
        """
        Parse object out of JSON data.

        Args:
            json: a dictionary.
        """
        validate(json, schema=cls._json_schema)

        c = cls(
            script=base64.b64decode(json["script"]),
            parameter_list=list(
                map(
                    lambda p: abi.ContractParameterDefinition.from_json(p),
                    json["parameters"],
                )
            ),
        )
        c.deployed = json["deployed"]

        return c

    def to_json(self) -> dict:
        """Convert object into JSON representation."""
        return {
            "script": base64.b64encode(self.script).decode("utf-8"),
            "parameters": list(
                map(
                    lambda index: {
                        "name": self.parameter_names[index],
                        "type": self.parameter_list[index].PascalCase(),
                    },
                    range(len(self.parameter_list)),
                )
            ),
            "deployed": self.deployed,
        }


class Account:
    """
    Container class for handling key material. Can be used to sign transactions.
    """

    _json_schema = {
        "type": "object",
        "properties": {
            "address": {"type": "string"},
            "label": {"type": ["string", "null"]},
            "isDefault": {"type": "boolean"},
            "lock": {"type": "boolean"},
            "key": {"type": ["string", "null"]},
            "contract": AccountContract._json_schema,
            "extra": {
                "type": ["object", "null"],
                "properties": {},
                "additionalProperties": True,
            },
        },
        "required": ["address", "label", "isDefault", "lock", "key", "extra"],
    }

    def __init__(
        self,
        private_key: Optional[bytes] = None,
        watch_only: bool = False,
        address: Optional[NeoAddress] = None,
        label: Optional[str] = None,
        lock: bool = False,
        contract_: Optional[contract.Contract] = None,
        extra: Optional[dict[str, Any]] = None,
        scrypt_parameters: Optional[scrypt.ScryptParameters] = None,
    ):
        """
        Do not call directly. Use the classmethods for creating an account.
        """
        public_key: Optional[cryptography.ECPoint] = None
        contract_script: Optional[bytes] = None

        if watch_only:
            if address is None:
                raise ValueError("Creating a watch only account requires an address")
            else:
                utils.validate_address(address)
        else:
            key_pair: cryptography.KeyPair

            if private_key is None:
                key_pair = cryptography.KeyPair.generate()
                private_key = key_pair.private_key
            else:
                key_pair = cryptography.KeyPair(private_key)
            contract_script = contractutils.create_signature_redeemscript(
                key_pair.public_key
            )
            script_hash = coreutils.to_script_hash(contract_script)
            address = address if address else utils.script_hash_to_address(script_hash)
            public_key = key_pair.public_key

        self.label: Optional[str] = label
        self.address: NeoAddress = address
        self.public_key = public_key
        self.lock = lock
        self.scrypt_parameters = scrypt_parameters
        self.private_key = private_key

        if isinstance(contract_, contract.Contract):
            contract_ = AccountContract.from_contract(contract_)
        elif contract_script is not None:
            default_parameters_list = [
                abi.ContractParameterDefinition(
                    name="signature", type=abi.ContractParameterType.SIGNATURE
                )
            ]
            contract_ = AccountContract(contract_script, default_parameters_list)
        self.contract: Optional[AccountContract] = contract_
        self.extra = extra if extra else {}

    def __eq__(self, other) -> bool:
        return isinstance(other, Account) and self.address == other.address

    @property
    def script_hash(self) -> types.UInt160:
        """
        Return account script hash.
        """
        return utils.address_to_script_hash(self.address)

    @property
    def is_watchonly(self) -> bool:
        """
        Return if the account can only be used for observing or in test transactions.
        """
        if self.private_key is None:
            return True
        else:
            return False

    @property
    def is_multisig(self) -> bool:
        """
        Return if the account requires multiple signers.
        """
        if self.contract is None:
            return False
        return contractutils.is_multisig_contract(self.contract.script)

    @property
    def is_single_sig(self) -> bool:
        """
        Return if the account requires a single signer.
        """
        if self.contract is None:
            return False
        return contractutils.is_signature_contract(self.contract.script)

    def add_as_sender(
        self,
        tx: transaction.Transaction,
        scope: Optional[verification.WitnessScope] = None,
    ):
        """
        Add the account as sender of the transaction.

        Args:
            tx: the transaction to modify.
            scope: the type of scope the signature of the sender has. Defaults to CALLED_BY_ENTRY.
             See Also: WitnessScope.
        """
        scope = scope if scope else verification.WitnessScope.CALLED_BY_ENTRY
        tx.signers.insert(0, verification.Signer(self.script_hash, scope))

    def sign_tx(self, tx: transaction.Transaction, magic: Optional[int] = None) -> None:
        """
        Helper function that signs the TX, adds the Witness and Sender.

        Args:
            tx: transaction to sign.
            magic: the network magic.

        Raises:
            ValueError: if transaction validation fails.
        """
        if magic is None:
            magic = settings.settings.network.magic

        self._validate_tx(tx)

        message = (
            magic.to_bytes(4, byteorder="little", signed=False) + tx.hash().to_array()
        )
        signature = self.sign(message)

        invocation_script = vm.ScriptBuilder().emit_push(signature).to_array()
        # mypy can't infer that the is_watchonly check ensures public_key has a value
        verification_script = contractutils.create_signature_redeemscript(self.public_key)  # type: ignore
        tx.witnesses.append(
            verification.Witness(invocation_script, verification_script)
        )

    def sign_multisig_tx(
        self,
        tx: transaction.Transaction,
        context: MultiSigContext,
        magic: Optional[int] = None,
    ) -> None:
        """
        Sign a transaction with a multi-signature account.

        Args:
            tx: the transaction to sign.
            context: the signing context.
            magic: override network magic.
        """
        if magic is None:
            magic = settings.settings.network.magic

        if not self.contract:
            raise ValueError("Account is not a valid multi-signature account")

        # When importing a multi-sig account it searches for an associated regular account to copy key material from.
        # However, it is possible to add a multi-sig account before having a regular account with one of the required
        # public keys for the multi-sig account. Therefore, we should check if we actually have key material to continue
        if self.is_watchonly:
            _, _, public_keys = contractutils.parse_as_multisig_contract(
                self.contract.script
            )
            raise ValueError(
                f"Cannot sign with watch only account. Try adding a regular account to your wallet "
                f"matching one of the following public keys, or update the key material for this account "
                f"directly."
                f" {list(map(lambda pk: str(pk), public_keys))}"
            )

        self._validate_tx(tx)

        if not self.is_multisig:
            raise ValueError("Account is not a valid multi-signature account")

        if not context.initialised:
            context.process_contract(self.contract.script)

        if self.public_key not in context.expected_public_keys:
            raise ValueError(
                "Account is not in the required key list for this signing context"
            )

        message = (
            magic.to_bytes(4, byteorder="little", signed=False) + tx.hash().to_array()
        )
        signature = self.sign(message)

        context.signature_pairs.update({self.public_key: signature})

        if context.is_complete:
            # build and insert multisig witness
            sb = vm.ScriptBuilder()

            pairs = list(context.signature_pairs.items())
            # sort by public key
            pairs.sort(key=lambda p: p[0])

            for i, (key, sig) in enumerate(pairs):
                if i == context.signing_threshold:
                    break
                sb.emit_push(sig)

            invocation_script = sb.to_array()
            verification_script = contractutils.create_multisig_redeemscript(
                context.signing_threshold, context.expected_public_keys
            )

            tx.witnesses.append(
                verification.Witness(invocation_script, verification_script)
            )

    def sign(self, data: bytes) -> bytes:
        """
        Sign arbitrary data using the SECP256R1 curve.

        Args:
            data: data to be signed.

        Returns:
            signature of the signed data.
        """
        if self.is_watchonly:
            raise ValueError("Cannot sign transaction using a watch only account")
        # mypy can't figure out that is_watchonly checks if private_key is None
        return cryptography.sign(data, self.private_key)  # type: ignore

    @classmethod
    def create_new(
        cls, scrypt_parameters: Optional[scrypt.ScryptParameters] = None
    ) -> Account:
        """
        Instantiate and returns a new account.

        Args:
            scrypt_parameters: supply custom Scrypt parameters.

        Returns:
            The newly created account.
        """
        return cls(watch_only=False, scrypt_parameters=scrypt_parameters)

    @classmethod
    def from_encrypted_key(
        cls,
        encrypted_key: str,
        password: str,
        scrypt_parameters: Optional[scrypt.ScryptParameters] = None,
    ) -> Account:
        """
        Instantiate and returns an account from a given key and password.
        Default settings assume a NEP-2 encrypted key.

        Args:
            encrypted_key: the encrypted private key.
            password: the password to decrypt the nep2 key.
            scrypt_parameters: supply custom Scrypt parameters.

        Returns:
            The newly created account.
        """
        return cls(
            private_key=cls.private_key_from_nep2(
                encrypted_key, password, scrypt_parameters
            ),
            scrypt_parameters=scrypt_parameters,
        )

    @classmethod
    def from_private_key(
        cls,
        private_key: bytes,
        scrypt_parameters: Optional[scrypt.ScryptParameters] = None,
    ) -> Account:
        """
        Instantiate and returns an account from a given private key and password.

        Args:
            private_key: the private key that will be used to create an encrypted key.
            scrypt_parameters: optional custom parameters to be used in the Scrypt algorithm. Default settings conform
            to the NEP-2 standard.

        Returns:
            the newly created account.
        """
        return cls(
            private_key=private_key,
            scrypt_parameters=scrypt_parameters,
        )

    @classmethod
    def from_wif(
        cls,
        wif: str,
        _scrypt_parameters: Optional[scrypt.ScryptParameters] = None,
    ) -> Account:
        """
        Instantiate and returns an account from a given wif and password.

        Args:
            wif: the wif that will be decrypted to get a private key and generate an encrypted key.
            _scrypt_parameters: the Scrypt parameters to use to encode the private key. Default conforms to NEP-2.

        Returns:
            the newly created account.
        """
        return cls(
            private_key=cls.private_key_from_wif(wif),
            scrypt_parameters=_scrypt_parameters,
        )

    @classmethod
    def watch_only(cls, script_hash: types.UInt160) -> Account:
        """
        Instantiate and returns a watch-only account from a given script hash.

        Args:
            script_hash: the script hash that will identify an account to be watched.

        Returns:
            the account that will be monitored.
        """
        return cls(
            watch_only=True,
            address=utils.script_hash_to_address(script_hash),
        )

    @classmethod
    def watch_only_from_address(cls, address: NeoAddress) -> Account:
        """
        Instantiate and returns a watch-only account from a given address.

        Args:
            address: the address that will identify an account to be watched.

        Returns:
            the account that will be monitored.
        """
        return cls(watch_only=True, address=address)

    def to_json(
        self, password: str, scrypt_parameters: Optional[scrypt.ScryptParameters] = None
    ) -> dict:
        """
        Encode account into NEP-6 JSON format

        Args:
             password: passphrase to use to encode the private key using NEP-2.
             scrypt_parameters: the Scrypt parameters to use to encode the private key. Default conforms to NEP-2.
        """
        return {
            "address": self.address,
            "label": self.label,
            "lock": self.lock,
            "key": (
                self.private_key_to_nep2(
                    self.private_key, password, scrypt_parameters
                ).decode("utf-8")
                if self.private_key is not None
                else None
            ),
            "contract": self.contract.to_json() if self.contract is not None else None,
            "extra": self.extra if len(self.extra) > 0 else None,
        }

    @classmethod
    def from_json(
        cls,
        json: dict,
        password: Optional[str],
        scrypt_parameters: Optional[scrypt.ScryptParameters] = None,
    ) -> Account:
        """
        Parse object out of JSON data.

        Args:
            json: a dictionary.
            password: the password to decrypt the json data.
            scrypt_parameters: the Scrypt parameters to use to encode the private key. Default conforms to NEP-2.

        Raises:
            KeyError: if the data supplied does not contain the necessary key.
        """
        validate(json, schema=cls._json_schema)

        private_key = None
        if json["key"] is not None and password is not None:
            private_key = cls.private_key_from_nep2(
                json["key"], password, scrypt_parameters
            )

        return cls(
            private_key=private_key,
            address=json["address"],
            label=json["label"],
            lock=json["lock"],
            contract_=AccountContract.from_json(json["contract"]),
            extra=json["extra"],
            scrypt_parameters=scrypt_parameters,
            watch_only=True if private_key is None else False,
        )

    @staticmethod
    def private_key_from_nep2(
        nep2_key: str,
        passphrase: str,
        _scrypt_parameters: Optional[scrypt.ScryptParameters] = None,
    ) -> bytes:
        """
        Decrypt a nep2 key into a private key.

        Args:
            nep2_key: the key that will be decrypted.
            passphrase: the password to decrypt the nep2 key.
            _scrypt_parameters: a ScryptParameters object that will be passed to the key derivation function.

        Raises:
            ValueError: if the length of the nep2_key is not valid.
            ValueError: if it's not possible to decode the nep2_key.
            ValueError: if the passphrase is incorrect or the version of the account is not valid.

        Returns:
            the private key.
        """
        if _scrypt_parameters is None:
            _scrypt_parameters = scrypt.ScryptParameters()

        if len(nep2_key) != 58:
            raise ValueError(
                f"Please provide a nep2_key with a length of 58 bytes (LEN: {len(nep2_key)})"
            )

        address_hash_size = 4
        address_hash_offset = len(NEP_FLAG) + len(NEP_HEADER)

        try:
            decoded_key = base58.b58decode_check(nep2_key)
        except Exception:
            raise ValueError("Base58decode failure of nep2 key")

        address_checksum = decoded_key[
            address_hash_offset : address_hash_offset + address_hash_size
        ]
        encrypted = decoded_key[-32:]

        pwd_normalized = bytes(unicodedata.normalize("NFC", passphrase), "utf-8")
        derived = hashlib.scrypt(
            password=pwd_normalized,
            salt=address_checksum,
            n=_scrypt_parameters.n,
            r=_scrypt_parameters.r,
            p=_scrypt_parameters.p,
            dklen=64,
        )

        derived1 = derived[:32]
        derived2 = derived[32:]

        cipher = AES.new(derived2, AES.MODE_ECB)
        decrypted = cipher.decrypt(encrypted)
        private_key = Account._xor_bytes(decrypted, derived1)

        # Now check that the address hashes match. If they don't, the password was wrong.
        key_pair = cryptography.KeyPair(private_key=private_key)
        script_hash = coreutils.to_script_hash(
            contractutils.create_signature_redeemscript(key_pair.public_key)
        )
        address = utils.script_hash_to_address(script_hash)
        first_hash = hashlib.sha256(address.encode("utf-8")).digest()
        second_hash = hashlib.sha256(first_hash).digest()
        checksum = second_hash[:4]
        if checksum != address_checksum:
            raise ValueError(
                f"Wrong passphrase or key was encrypted with an address version that is not "
                f"{settings.settings.network.account_version}"
            )

        return private_key

    @staticmethod
    def private_key_to_nep2(
        private_key: bytes,
        passphrase: str,
        _scrypt_parameters: Optional[scrypt.ScryptParameters] = None,
    ) -> bytes:
        """
        Encrypt a private key into a nep2 key.

        Args:
            private_key: the key that will be encrypted.
            passphrase: the password to encrypt the nep2 key.
            _scrypt_parameters: a ScryptParameters object that will be passed to the key derivation function.

        Returns:
            the encrypted nep2 key.
        """
        if _scrypt_parameters is None:
            _scrypt_parameters = scrypt.ScryptParameters()

        key_pair = cryptography.KeyPair(private_key=private_key)
        script_hash = coreutils.to_script_hash(
            contractutils.create_signature_redeemscript(key_pair.public_key)
        )
        address = utils.script_hash_to_address(script_hash)
        # NEP2 checksum: hash the address twice and get the first 4 bytes
        first_hash = hashlib.sha256(address.encode("utf-8")).digest()
        second_hash = hashlib.sha256(first_hash).digest()
        checksum = second_hash[:4]

        pwd_normalized = bytes(unicodedata.normalize("NFC", passphrase), "utf-8")
        derived = hashlib.scrypt(
            password=pwd_normalized,
            salt=checksum,
            n=_scrypt_parameters.n,
            r=_scrypt_parameters.r,
            p=_scrypt_parameters.p,
            dklen=64,
        )

        derived1 = derived[:32]
        derived2 = derived[32:]

        xor_ed = Account._xor_bytes(bytes(private_key), derived1)
        cipher = AES.new(derived2, AES.MODE_ECB)
        encrypted = cipher.encrypt(xor_ed)

        nep2 = bytearray()
        nep2.extend(NEP_HEADER)
        nep2.extend(NEP_FLAG)
        nep2.extend(checksum)
        nep2.extend(encrypted)

        # Finally, encode with Base58Check
        encoded_nep2 = base58.b58encode_check(bytes(nep2))

        return encoded_nep2

    @staticmethod
    def private_key_from_wif(wif: str) -> bytes:
        """
        Decrypt a private key from a wif.

        Args:
            wif: the wif that will be decrypted.

        Raises:
            ValueError: if the wif is not valid.
        """
        try:
            decoded_key: bytes = base58.b58decode_check(wif)
        except Exception:
            raise ValueError("Base58decode failure of wif")

        if len(decoded_key) != 34:
            raise ValueError(
                f"The decoded wif length should be "
                f"{len(WIF_PREFIX) + PRIVATE_KEY_LENGTH + len(WIF_SUFFIX)}, while the given wif "
                f"length is {len(decoded_key)}"
            )
        elif decoded_key[:1] != WIF_PREFIX:
            raise ValueError(f"The decoded wif first byte should be {str(WIF_PREFIX)}")
        elif decoded_key[-1:] != WIF_SUFFIX:
            raise ValueError(f"The decoded wif last byte should be {str(WIF_SUFFIX)}")

        private_key = decoded_key[1:33]

        return private_key

    @staticmethod
    def private_key_to_wif(private_key: bytes) -> str:
        return base58.b58encode_check(WIF_PREFIX + private_key + WIF_SUFFIX).decode()

    @staticmethod
    def _xor_bytes(a: bytes, b: bytes) -> bytes:
        """
        XOR on two bytes objects

        Args:
            a (bytes): object 1
            b (bytes): object 2
        Returns:
            the XOR result
        """
        assert len(a) == len(b)
        res = bytearray()
        for i in range(len(a)):
            res.append(a[i] ^ b[i])
        return bytes(res)

    def _validate_tx(self, tx: transaction.Transaction) -> None:
        """
        Helper to validate properties before signing
        """
        if tx.network_fee == 0 or tx.system_fee == 0:
            raise ValueError(
                "Transaction validation failure - "
                "a transaction without network and system fees will always fail to validate on chain"
            )

        if len(tx.signers) == 0:
            raise ValueError("Transaction validation failure - Missing sender")

        if len(tx.script) == 0:
            raise ValueError(
                "Transaction validation failure - script field can't be empty"
            )

        if self.is_watchonly:
            raise ValueError("Cannot sign transaction using a watch only account")



"""
Builder for creating a Transaction. Calculate fees, add signers and sign.
"""

from neo3.api import noderpc
from neo3.api.helpers import signing
from neo3.network.payloads import transaction, verification
from neo3.wallet import account
from typing import Optional


class TxBuilder:
    """
    Transaction builder.
    """

    def __init__(self, client: noderpc.NeoRpcClient, script: Optional[bytes] = None):
        self.client = client
        self.tx = transaction.Transaction(
            version=0,
            nonce=123,
            system_fee=0,
            network_fee=0,
            valid_until_block=0,
            attributes=[],
            signers=[],
            script=b"" if script is None else script,
        )
        self.signing_funcs: list[signing.SigningFunction] = []
        self.network = -1

    async def init(self) -> None:
        """
        Initialize the builder.
        """
        res = await self.client.get_version()
        self.network = res.protocol.network

    async def calculate_system_fee(self) -> None:
        """
        Calculates and set the system fee. Requires at least one signer.
        """
        if len(self.tx.signers) == 0:
            raise ValueError(
                "Need at least one signer (a.k.a the sender who pays for the transaction) or the "
                "fee calculation will be incorrect"
            )
        res = await self.client.invoke_script(self.tx.script, self.tx.signers)
        self.tx.system_fee = res.gas_consumed

    async def set_valid_until_block(self, blocks: int = 1500) -> None:
        """
        Set maximum time the transaction is valid in the mempool. Defaults to about 24h for a network with 15s blocktime.

        Args:
            blocks: until how many blocks from the current chain height is the transaction valid. Defaults to ~24 hours.
        """
        self.tx.valid_until_block = await self.client.get_block_count() + blocks

    async def calculate_network_fee(self) -> None:
        """
        Calculate and set the network fee. Requires at least one signer.
        """
        if len(self.tx.witnesses) == 0:
            if len(self.tx.signers) == 0:
                raise ValueError("Cannot calculate network fee without signers")
            # adding a witness(es) so we can calculate the network fee
            for _ in range(len(self.tx.signers)):
                self.tx.witnesses.append(TxBuilder._dummy_signing_witness())
            self.tx.network_fee = await self.client.calculate_network_fee(self.tx)
            # removing it here as it will be replaced by a proper one once we're signing
            self.tx.witnesses = []
        else:
            if len(self.tx.signers) == 0:
                raise ValueError("Cannot calculate network fee without signers")
            self.tx.network_fee = await self.client.calculate_network_fee(self.tx)

    @staticmethod
    def _dummy_signing_witness() -> verification.Witness:
        """single signature account witness"""
        acc = account.Account.create_new()
        if acc.contract is None:
            raise Exception(
                "Unreachable"
            )  # we know this can't happen, but mypy doesn't
        return verification.Witness(
            invocation_script=b"", verification_script=acc.contract.script
        )

    async def build_and_sign(self) -> transaction.Transaction:
        """
        Sign the transaction with all signers and return the finalized result.
        """
        len_signers = len(self.tx.signers)
        if len_signers == 0:
            raise ValueError("Cannot sign transaction without signers")

        if self.network == -1:
            raise ValueError(
                "Network value not valid (-1). Call init() to automatically sync it from the network or set the `network` attribute"
            )

        for f, s in zip(self.signing_funcs, self.tx.signers):
            await f(self.tx, signing.SigningDetails(self.network))
        return self.tx

    def build_unsigned(self) -> transaction.Transaction:
        """
        Return the unsigned transaction. For example for use in an offline signing scenario.
        """
        return self.tx

    def add_signer(
        self, func: signing.SigningFunction, signer: verification.Signer
    ) -> None:
        """
        Add a Signer with scopes to the transaction and its signing function.

        Args:
            func: one of neo3.api.helpers.signing.
            signer: a Signer determining the validity of the signature.

        Returns:

        """
        for s in self.tx.signers:
            if signer.account == s.account:
                raise ValueError(
                    f"Signer with same account ({signer.account} already exists."
                )
        self.tx.signers.append(signer)
        self.signing_funcs.append(func)



import socket
import logging
import asyncio
import unittest
import sys

from neo3.network import node, message, capabilities, ipfilter
from neo3.network.payloads import (
    version,
    address,
    block,
    empty,
    inventory,
    ping,
    transaction,
)
from copy import deepcopy
from neo3 import network_logger
from neo3.settings import settings
from neo3.core import types
from unittest import mock, IsolatedAsyncioTestCase
from tests import helpers as test_helpers
import platform


class NeoNodeTestCase(IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        if platform.system() == "Windows" and (
            platform.python_version_tuple()[1] == "12"
            or platform.python_version_tuple()[1] == "13"
        ):
            raise unittest.SkipTest(
                "skipping these tests bcause something with socketpair is failing for 3.12 and 3.13\ninvestigate later"
            )
        # network_logger = logging.getLogger("neo3.network")
        # network_logger.setLevel(logging.DEBUG)
        # stdio_handler = logging.StreamHandler()
        # network_logger.addHandler(stdio_handler)

        caps = [
            capabilities.FullNodeCapability(0),
            capabilities.ServerCapability(
                n_type=capabilities.NodeCapabilityType.TCPSERVER, port=10333
            ),
        ]

        cls.m_version = message.Message(
            msg_type=message.MessageType.VERSION,
            payload=version.VersionPayload(1, "NEO3-MOCK-CLIENT", caps),
        )
        cls.m_verack = message.Message(msg_type=message.MessageType.VERACK)

        host = "127.0.0.1"
        port = 1111
        cls.peername_data = {"peername": (host, port)}

    @classmethod
    def tearDownClass(cls) -> None:
        settings.reset_settings_to_default()
        network_logger = logging.getLogger("neo3.network")
        network_logger.handlers.clear()

    def setUp(self) -> None:
        settings.reset_settings_to_default()
        node.NeoNode._reset_for_test()
        ipfilter.ipfilter.reset()

    async def test_connect_to(self):
        with self.assertRaises(ValueError) as context:
            await node.NeoNode.connect_to()
        self.assertIn(
            "host and port was not specified and no sock specified",
            str(context.exception),
        )

        with self.assertRaises(ValueError) as context:
            await node.NeoNode.connect_to(host="123", socket=object())
        self.assertIn(
            "host/port and socket can not be specified at the same time",
            str(context.exception),
        )

    async def test_connect_to_with_socket(self):
        r, w = socket.socketpair()
        loop = asyncio.get_running_loop()

        host = "127.0.0.1"
        port = 1111
        test_data = {"peername": (host, port)}

        with self.assertLogs(network_logger, "DEBUG") as log_context:
            # hand-shake data
            loop.call_soon(w.send, self.m_version.to_array())
            loop.call_soon(w.send, self.m_verack.to_array())
            n, _ = await node.NeoNode.connect_to(socket=r, _test_data=test_data)
        r.close()
        w.close()
        self.assertIn("Trying to connect to socket", log_context.output[0])
        self.assertIn(
            f"Connected to NEO3-MOCK-CLIENT @ {host}:{port}: 0", log_context.output[2]
        )
        self.assertIsInstance(n, node.NeoNode)
        await n.disconnect(address.DisconnectReason.SHUTTING_DOWN)

    async def test_connect_to_with_host_ip(self):
        # just for basic coverage, we use the above socket mock for more indepth testing
        host = "127.0.0.1"
        port = 1111

        with self.assertLogs(network_logger, "DEBUG") as log_context:
            with mock.patch.object(asyncio, "open_connection"):
                _, _ = await node.NeoNode.connect_to(host, port)
        self.assertIn(f"Trying to connect to: {host}:{port}", log_context.output[0])

    async def test_connect_to_exceptions(self):
        with mock.patch.object(asyncio, "open_connection") as mocked_open_conn:
            mocked_open_conn.side_effect = asyncio.TimeoutError
            node_instance, failure = await node.NeoNode.connect_to(socket=object())
            self.assertIsNone(node_instance)
            self.assertEqual("Timed out", failure[1])

        with mock.patch.object(asyncio, "open_connection") as mocked_open_conn:
            mocked_open_conn.side_effect = OSError("unreachable")
            node_instance, failure = await node.NeoNode.connect_to(socket=object())
            self.assertIsNone(node_instance)
            self.assertEqual("Failed to connect for reason unreachable", failure[1])

        with mock.patch.object(asyncio, "open_connection") as mocked_open_conn:
            mocked_open_conn.side_effect = asyncio.CancelledError
            node_instance, failure = await node.NeoNode.connect_to(socket=object())
            self.assertIsNone(node_instance)
            self.assertEqual("Cancelled", failure[1])

    def test_helpers(self):
        addr1 = address.NetworkAddress(address="127.0.0.1:1111")
        addr1.set_state_dead()
        addr2 = address.NetworkAddress(address="127.0.0.1:2222")
        node.NeoNode.addresses = [addr1, addr2]
        result = node.NeoNode.get_address_new()
        self.assertEqual(addr2, result)

        n = node.NeoNode(object(), object())
        addr = n._find_address_by_host_port("127.0.0.1:1111")
        self.assertEqual(addr1, addr)
        addr = n._find_address_by_host_port("127.0.0.1:3333")
        self.assertEqual(None, addr)

    async def test_handshake_first_message_not_VERSION(self):
        r, w = socket.socketpair()
        loop = asyncio.get_running_loop()

        with self.assertLogs(network_logger, "DEBUG") as log_context:
            loop.call_soon(w.send, self.m_verack.to_array())
            await node.NeoNode.connect_to(socket=r, _test_data=self.peername_data)
        r.close()
        w.close()
        self.assertIn(
            "Disconnect called with reason=HANDSHAKE_VERSION_ERROR",
            log_context.output[1],
        )

    async def test_handshake_version_validation_failed(self):
        # mock the version validation call result. The version_validation function is tested down below
        settings.network.magic = 769
        r, w = socket.socketpair()
        loop = asyncio.get_running_loop()
        version = deepcopy(self.m_version)
        # ensure the network magic read from the stream is different
        version.payload.magic = 123
        with self.assertLogs(network_logger, "DEBUG") as log_context:
            loop.call_soon(w.send, version.to_array())
            await node.NeoNode.connect_to(socket=r, _test_data=self.peername_data)
        r.close()
        w.close()
        self.assertIn(
            "Disconnect called with reason=HANDSHAKE_VERSION_ERROR",
            log_context.output[2],
        )

    async def test_handshake_second_message_not_VERACK(self):
        # ensure we have the same magic
        settings.network.magic = self.m_version.payload.magic

        r, w = socket.socketpair()
        loop = asyncio.get_running_loop()

        with self.assertLogs(network_logger, "DEBUG") as log_context:
            loop.call_soon(w.send, self.m_version.to_array())
            loop.call_soon(w.send, self.m_version.to_array())
            await node.NeoNode.connect_to(socket=r, _test_data=self.peername_data)
        r.close()
        w.close()
        self.assertIn(
            "Disconnect called with reason=HANDSHAKE_VERACK_ERROR",
            log_context.output[2],
        )

    def _new_version(self):
        self.start_height = 888
        caps = [
            capabilities.FullNodeCapability(start_height=self.start_height),
            capabilities.ServerCapability(
                capabilities.NodeCapabilityType.TCPSERVER, 10333
            ),
        ]
        return version.VersionPayload(
            nonce=123, user_agent="NEO3-MOCK-CLIENT", capabilities=caps
        )

    def test_version_validation_client_is_self(self):
        n = node.NeoNode(object(), object())

        # test for client is self
        version = self._new_version()
        version.nonce = n.nodeid
        with self.assertLogs(network_logger, "DEBUG") as log_context:
            result = n._validate_version(version)
        self.assertFalse(result)
        self.assertIn("Client is self", log_context.output[0])

    def test_version_validation_wrong_network(self):
        n = node.NeoNode(object(), object())

        # test for wrong network
        settings.network.magic = 769
        version = self._new_version()
        version.magic = 111
        with self.assertLogs(network_logger, "DEBUG") as log_context:
            result = n._validate_version(version)
        self.assertFalse(result)
        self.assertIn("Wrong network id", log_context.output[0])

    def test_version_validation_should_updating_address_to_connected_state(self):
        n = node.NeoNode(object(), object())

        # test updating address state to CONNECTED
        # this is relevant for addresses that have been added through the nodemanager based on the seedlist
        # set the addresses ourselves to mimick the nodemanager startup behaviour
        node.NeoNode.addresses = [n.address]
        version = self._new_version()

        result = n._validate_version(version)
        # validate our address is now set to CONNECTED
        self.assertTrue(result)
        self.assertTrue(n.address.is_state_connected)
        self.assertEqual(self.start_height, n.best_height)

    def test_version_validation_should_add_new_address(self):
        # when using the `connect_to` method or when a server is hosted accepting incoming clients
        # we should add the address to our know addresses list
        n = node.NeoNode(object(), object())
        # normally this is updated by `connection_made()`, since we skip that in this test we set it manually
        n.address.address = "127.0.0.1:1111"
        self.assertEqual(0, len(node.NeoNode.addresses))

        version = self._new_version()
        with self.assertLogs(network_logger, "DEBUG") as log_context:
            n._validate_version(version)
        self.assertEqual(1, len(node.NeoNode.addresses))
        self.assertIn(
            "Adding address from outside 127.0.0.1:1111", log_context.output[0]
        )

    def test_version_validation_fail_if_no_full_node_capabilities(self):
        n = node.NeoNode(object(), object())

        version = self._new_version()
        # remove the full node capability
        version.capabilities.pop(0)
        result = n._validate_version(version)
        self.assertFalse(result)

    @unittest.skipIf(
        sys.platform.startswith("win") and sys.version_info >= (3, 14, 0),
        "Skip this test on Windows for Python 3.14+",
    )
    async def test_connect_blocked_by_ipfilter(self):
        ipfilter.ipfilter.blacklist_add("127.0.0.1")
        r, w = socket.socketpair()
        loop = asyncio.get_running_loop()
        with self.assertLogs(network_logger, "DEBUG") as log_context:
            loop.call_soon(w.send, self.m_version.to_array())
            loop.call_soon(w.send, self.m_verack.to_array())
            await node.NeoNode.connect_to(socket=r, _test_data=self.peername_data)
        r.close()
        w.close()
        self.assertIn("Blocked by ipfilter: 127.0.0.1:1111", log_context.output[1])

    async def test_req_addr_list(self):
        n = node.NeoNode(object(), object())
        n.send_message = mock.AsyncMock()
        await n.request_address_list()

        self.assertIsNotNone(n.send_message.call_args)
        m = n.send_message.call_args[0][0]  # type: message.Message
        self.assertEqual(message.MessageType.GETADDR, m.type)
        self.assertIsInstance(m.payload, empty.EmptyPayload)

    async def test_send_addr_list(self):
        n = node.NeoNode(object(), object())
        n.send_message = mock.AsyncMock()
        n.addresses = [address.NetworkAddress(address="127.0.0.1:1111")]
        await n.send_address_list(n.addresses)

        self.assertIsNotNone(n.send_message.call_args)
        m = n.send_message.call_args[0][0]  # type: message.Message
        self.assertEqual(message.MessageType.ADDR, m.type)
        self.assertIsInstance(m.payload, address.AddrPayload)
        self.assertEqual(n.addresses, m.payload.addresses)

    async def test_req_headers(self):
        n = node.NeoNode(object(), object())
        n.send_message = mock.AsyncMock()
        index_start = 0
        count = 10

        await n.request_headers(index_start, count)
        self.assertIsNotNone(n.send_message.call_args)
        m = n.send_message.call_args[0][0]  # type: message.Message
        self.assertEqual(message.MessageType.GETHEADERS, m.type)
        self.assertIsInstance(m.payload, block.GetBlockByIndexPayload)
        self.assertEqual(index_start, m.payload.index_start)
        self.assertEqual(count, m.payload.count)

    async def test_send_headers(self):
        n = node.NeoNode(object(), object())
        n.send_message = mock.AsyncMock()
        headers = 2001 * [test_helpers.SerializableObject()]

        await n.send_headers(headers)
        self.assertIsNotNone(n.send_message.call_args)
        m = n.send_message.call_args[0][0]  # type: message.Message
        self.assertEqual(message.MessageType.HEADERS, m.type)
        self.assertIsInstance(m.payload, block.HeadersPayload)
        # max sure it clips the size to max 2K headers
        self.assertEqual(2000, len(m.payload.headers))

    async def test_req_blocks(self):
        n = node.NeoNode(object(), object())
        n.send_message = mock.AsyncMock()
        hash_start = types.UInt256.from_string(
            "65793a030c0dcd4fff4da8a6a6d5daa8b570750da4fdeea1bbc43bdf124aedc9"
        )
        count = 10

        await n.request_blocks(hash_start, count)
        self.assertIsNotNone(n.send_message.call_args)
        m = n.send_message.call_args[0][0]  # type: message.Message
        self.assertEqual(message.MessageType.GETBLOCKS, m.type)
        self.assertIsInstance(m.payload, block.GetBlocksPayload)
        self.assertEqual(hash_start, m.payload.hash_start)
        self.assertEqual(count, m.payload.count)

    async def test_req_block_data(self):
        n = node.NeoNode(object(), object())
        n.send_message = mock.AsyncMock()
        index_start = 1
        count = 2

        await n.request_block_data(index_start, count)
        self.assertIsNotNone(n.send_message.call_args)
        m = n.send_message.call_args[0][0]  # type: message.Message
        self.assertEqual(message.MessageType.GETBLOCKBYINDEX, m.type)
        self.assertIsInstance(m.payload, block.GetBlockByIndexPayload)
        self.assertEqual(index_start, m.payload.index_start)
        self.assertEqual(count, m.payload.count)

    async def test_request_data(self):
        n = node.NeoNode(object(), object())
        n.send_message = mock.AsyncMock()
        hash1 = types.UInt256.from_string(
            "65793a030c0dcd4fff4da8a6a6d5daa8b570750da4fdeea1bbc43bdf124aedc9"
        )
        hash2 = types.UInt256.from_string(
            "65793a030c0dcd4fff4da8a6a6d5daa8b570750da4fdeea1bbc43bdf124aaaaa"
        )
        hashes = [hash1, hash2]

        await n.request_data(inventory.InventoryType.BLOCK, hashes)
        self.assertIsNotNone(n.send_message.call_args)
        m = n.send_message.call_args[0][0]  # type: message.Message
        self.assertEqual(message.MessageType.GETDATA, m.type)
        self.assertIsInstance(m.payload, inventory.InventoryPayload)
        self.assertEqual(hashes, m.payload.hashes)

    async def test_send_inventory_and_relay(self):
        # test 2 in 1
        # taken from the Transaction testcase in `test_payloads.py`
        raw_tx = bytes.fromhex(
            "007B000000C8010000000000001503000000000000010000000154A64CAC1B1073E662933EF3E30B007CD98D67D7000002010201000155"
        )
        tx = transaction.Transaction.deserialize_from_bytes(raw_tx)

        n = node.NeoNode(object(), object())
        n.send_message = mock.AsyncMock()

        await n.relay(tx)
        m = n.send_message.call_args[0][0]  # type: message.Message
        self.assertEqual(message.MessageType.INV, m.type)
        self.assertIsInstance(m.payload, inventory.InventoryPayload)
        self.assertEqual(tx.hash(), m.payload.hashes[0])

    async def test_processing_messages(self):
        m_addr = message.Message(
            msg_type=message.MessageType.ADDR, payload=address.AddrPayload([])
        )
        m_block = message.Message(
            msg_type=message.MessageType.BLOCK, payload=block.Block._serializable_init()
        )
        m_inv1 = message.Message(
            msg_type=message.MessageType.INV,
            payload=inventory.InventoryPayload(
                inventory.InventoryType.BLOCK,
                hashes=[
                    types.UInt256.from_string(
                        "65793a030c0dcd4fff4da8a6a6d5daa8b570750da4fdeea1bbc43bdf124aedc9"
                    )
                ],
            ),
        )
        m_inv2 = message.Message(
            msg_type=message.MessageType.INV,
            payload=inventory.InventoryPayload(inventory.InventoryType.TX, []),
        )
        m_getaddr = message.Message(
            msg_type=message.MessageType.GETADDR, payload=empty.EmptyPayload()
        )
        m_mempool = message.Message(
            msg_type=message.MessageType.MEMPOOL, payload=empty.EmptyPayload()
        )

        # taken from the Headers testcase in `test_payloads`
        raw_headers_payload = bytes.fromhex(
            "0000000001FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00A402FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00A400000000000000007B00000000F7B4D00143932F3B6243CFC06CB4A68F22C739E201020102020304"
        )
        m_headers = message.Message(
            msg_type=message.MessageType.HEADERS,
            payload=block.HeadersPayload.deserialize_from_bytes(raw_headers_payload),
        )
        m_ping = message.Message(
            msg_type=message.MessageType.PING, payload=ping.PingPayload(0)
        )
        m_pong = message.Message(
            msg_type=message.MessageType.PONG, payload=ping.PingPayload(0)
        )
        m_reject = message.Message(
            msg_type=message.MessageType.REJECT, payload=empty.EmptyPayload()
        )

        def _mock_data(self):
            # first do handshake
            yield self.m_version.to_array()
            yield self.m_verack.to_array()
            # next send all types of messages we handle
            yield m_addr.to_array()
            yield m_block.to_array()
            yield m_inv1.to_array()
            yield m_inv2.to_array()
            yield m_getaddr.to_array()
            yield m_mempool.to_array()
            yield m_headers.to_array()
            yield m_ping.to_array()
            yield m_pong.to_array()
            yield m_reject.to_array()

        loop = asyncio.get_running_loop()
        r, w = socket.socketpair()

        data = b"".join(list(_mock_data(self)))
        with self.assertLogs(network_logger, "DEBUG") as log_context:
            try:
                loop.call_soon(w.send, data)
                n, _ = await node.NeoNode.connect_to(
                    socket=r, _test_data=self.peername_data
                )
                n.start_message_handler()
            except Exception as e:
                print(f"Unexpected: {e}")

            await asyncio.sleep(0.5)
            await n.disconnect(address.DisconnectReason.SHUTTING_DOWN)
        r.close()
        w.close()

    def test_utility_function(self):
        with self.assertRaises(ValueError) as context:
            node.encode_base62(-100)
        self.assertIn("cannot encode negative numbers", str(context.exception))

        self.assertEqual("0", node.encode_base62(0))

        self.assertEqual("w1R", node.encode_base62(123123))



import abc


class IJson(abc.ABC):
    @abc.abstractmethod
    def to_json(self) -> dict:
        """Convert object into JSON representation."""

    @classmethod
    @abc.abstractmethod
    def from_json(cls, json: dict):
        """Create object from JSON"""



from enum import IntFlag


class FindOptions(IntFlag):
    """
    Possible search options when using the `System.Storage.Find` SYSCALL.
    """

    NONE = 0
    KEYS_ONLY = 1 << 0
    REMOVE_PREFIX = 1 << 1
    VALUES_ONLY = 1 << 2
    DESERIALIZE_VALUES = 1 << 3
    PICK_FIELD0 = 1 << 4
    PICK_FIELD1 = 1 << 5
    BACKWARDS = 1 << 7
    ALL = (
        KEYS_ONLY
        | REMOVE_PREFIX
        | VALUES_ONLY
        | DESERIALIZE_VALUES
        | PICK_FIELD0
        | PICK_FIELD1
        | BACKWARDS
    )



import unittest
import base64
from neo3.network.payloads.transaction import Transaction


class TestPayloadIssues(unittest.TestCase):
    def shortDescription(self):
        # disable docstring printing in test runner
        return None

    def test_witnesscondition_deserialization_of_tx_in_block_3367174(self):
        # ref: https://app.clickup.com/t/861mv391t
        # should not raise a ValueError
        try:
            raw_tx = base64.b64decode(
                "AE02bF6kD0ABAAAAAMjsAQAAAAAAJGEzAAG2cn7NBD7czWw0a+Oa4gSfyK4ROkABAQMCIBh8ns+7w1/Ttac3ifPmgPd4UTc4uQBeEQMtL9OkMAAAAAwUl6Y6Dqqt8iuCFxjZeiafFLqIGLkMFLZyfs0EPtzNbDRr45riBJ/IrhE6FMAfDAh0cmFuc2ZlcgwU6aNc0l5Z4RDGaj8d715RgxKfBJtBYn1bUgFCDEBEnZaxHmL9kNpS5XlIjdiJwiJWgw0GB9O1D4y0sGRe+rSaO0fZzaN2VfkJnxqR60oVEuf8xiE7cZD28cDifLUXKAwhA5FClptV/FtBBClyJ1f/+HKtWWUXMw6H3NcOZdzd2gcuQVbnsyc="
            )
            tx = Transaction.deserialize_from_bytes(raw_tx)
        except ValueError as e:
            if "Deserialization error - unknown witness condition" in str(e):
                self.assertTrue(False)



from __future__ import annotations
from .merkletree import MerkleTree
from .bloomfilter import BloomFilter
from .ecc import ECCCurve, ECPoint, KeyPair, ECCException, ecdsa_verify, ecdsa_sign
import hashlib

__all__ = [
    "MerkleTree",
    "BloomFilter",
    "ECCCurve",
    "ECPoint",
    "KeyPair",
    "sign",
    "verify_signature",
    "ECCException",
]


def sign(
    message: bytes,
    private_key: bytes,
    curve=ECCCurve.SECP256R1,
    hash_func=hashlib.sha256,
) -> bytes:
    return ecdsa_sign(private_key, message, curve, hash_func)


def verify_signature(
    message: bytes,
    signature: bytes,
    public_key: bytes,
    curve: ECCCurve = ECCCurve.SECP256R1,
    hash_func=hashlib.sha256,
) -> bool:
    """
    Test is the `signature` is signed by `public_key` valid for `message`.

    Args:
        message: the data the signature was created over
        signature: the signature to validate for `message`
        public_key: the public key the message was signed with
        curve: the ECC curve to use for verifying

    Raises:
        ValueError: for the Secp256r1 curve if the public key has an invalid format

    """
    pub_key = ECPoint(public_key, curve, True)
    return ecdsa_verify(signature, message, pub_key, hash_func)



import pathlib
import asyncio
from neo3.wallet.wallet import Wallet, account
from neo3.core import types
from neo3.api.wrappers import GenericContract, NeoToken, GasToken
from neo3.api.helpers.signing import sign_with_account, sign_with_multisig_account
from neo3.contracts import nef, manifest
from neo3.network.payloads.verification import Signer
from boaconstructor import NeoGoNode

shared_dir = pathlib.Path("shared").resolve(strict=True)

user_wallet = Wallet.from_file(f"{shared_dir}/user-wallet.json", passwords=["123"])
coz_wallet = Wallet.from_file(f"{shared_dir}/coz-wallet.json", passwords=["123"])

coz_token_hash = types.UInt160.from_string("0x41ee5befd936c90f15893261abbd681f20ed0429")
# corresponds to the nep-11 token in the `/nep11-token/` dir and deployed with the `coz` account
nep11_token_hash = types.UInt160.from_string(
    "0x35de2913c480c19a7667da1cc3b2fe3e4c9de761"
)


class ExampleNode(NeoGoNode):
    @property
    def rpc_host(self) -> str:
        return self.facade.rpc_host

    def __enter__(self):
        self.start()
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self._setup_for_test())
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    async def _setup_for_test(self):
        sign_pair = (
            sign_with_multisig_account(self.account_committee),
            Signer(self.account_committee.script_hash),
        )
        neo = NeoToken()
        gas = GasToken()
        coz_account = coz_wallet.account_default
        user_account = user_wallet.account_default
        await self.facade.invoke(
            neo.transfer_friendly(
                self.account_committee.script_hash, coz_account.script_hash, 50000000, 0
            ),
            signers=[sign_pair],
        )
        await self.facade.invoke(
            gas.transfer_friendly(
                self.account_committee.script_hash, coz_account.script_hash, 26000000, 8
            ),
            signers=[sign_pair],
        )
        await self.facade.invoke(
            neo.transfer_friendly(
                self.account_committee.script_hash,
                user_account.script_hash,
                50000000,
                0,
            ),
            signers=[sign_pair],
        )
        await self.facade.invoke(
            gas.transfer_friendly(
                self.account_committee.script_hash,
                user_account.script_hash,
                26000000,
                8,
            ),
            signers=[sign_pair],
        )

        await self._deploy_contract(
            f"{shared_dir}/nep17-token/nep17token.nef", coz_account
        )
        await self._deploy_contract(
            f"{shared_dir}/nep11-token/nep11-token.nef", coz_account
        )

        sign_with_user = (
            sign_with_account(user_account),
            Signer(user_account.script_hash),
        )

        await self.facade.invoke(
            neo.transfer_friendly(
                user_account.script_hash, "NgJ6aLeAi3wJAQ3JbgcWsHGwUT76bvcWMM", 100, 0
            ),
            signers=[sign_with_user],
        )

        sign_with_coz = (
            sign_with_account(coz_account),
            Signer(coz_account.script_hash),
        )
        await self.facade.invoke(
            neo.candidate_register(coz_account.public_key), signers=[sign_with_coz]
        )

    async def _deploy_contract(
        self, nef_path: str, signing_account: account.Account
    ) -> types.UInt160:
        _nef = nef.NEF.from_file(nef_path)
        manifest_path = nef_path.removesuffix(".nef") + ".manifest.json"
        _manifest = manifest.ContractManifest.from_file(manifest_path)

        if signing_account.is_multisig:
            sign_pair = (
                sign_with_multisig_account(signing_account),
                Signer(signing_account.script_hash),
            )
        else:
            sign_pair = (
                sign_with_account(signing_account),
                Signer(signing_account.script_hash),
            )

        receipt = await self.facade.invoke(
            GenericContract.deploy(_nef, _manifest), signers=[sign_pair]
        )
        return receipt.result



"""
Classes for managing transaction signers and signature scopes.
"""

from __future__ import annotations
import hashlib
import abc
import base64
from enum import IntFlag, IntEnum
from neo3.core import serialization, utils, types, cryptography, Size as s, interfaces
from typing import Optional, Any, no_type_check, Iterator
from collections.abc import Sequence


class WitnessScope(IntFlag):
    """
    Flags that determine where in the system the signature is valid. Used by `CheckWitness()` sys call.
    """

    #: No Contract was witnessed. Only sign the transaction.
    NONE = 0x0
    #: Allow the witness if the current calling script hash equals the entry script hash into the virtual machine.
    #: Using this prevents passing `CheckWitness()` in a smart contract called via another smart contract.
    CALLED_BY_ENTRY = 0x01
    #: Allow the witness if called from a smart contract that is whitelisted in the signer `allowed_contracts`
    #: attribute.
    CUSTOM_CONTRACTS = 0x10
    #: Allow the witness if any public key is in the signer `allowed_groups` attribute is whitelisted in the contracts
    #: manifest.groups array.
    CUSTOM_GROUPS = 0x20
    #: Allow the witness if the specified `Signer.rules` are satisfied
    WITNESS_RULES = 0x40
    #: Allow the witness in all context. Equal to NEO 2.x's default behaviour.
    GLOBAL = 0x80

    @no_type_check
    def to_csharp_name(self) -> str:
        #: Internal helper to match C# convention.
        if self == 0:
            return "None"
        flags = []
        if self.CALLED_BY_ENTRY in self:
            flags.append("CalledByEntry")
        if self.CUSTOM_CONTRACTS in self:
            flags.append("CustomContracts")
        if self.CUSTOM_GROUPS in self:
            flags.append("CustomGroups")
        if self.GLOBAL in self:
            flags.append("Global")
        if self.WITNESS_RULES in self:
            flags.append("WitnessRules")
        return ", ".join(flags)

    @classmethod
    def from_chsarp_name(cls, csharp_name):
        #: Internal helper to parse from C# convention
        c = cls(cls.NONE)
        if "CalledByEntry" in csharp_name:
            c |= c.CALLED_BY_ENTRY
        if "CustomContracts" in csharp_name:
            c |= c.CUSTOM_CONTRACTS
        if "CustomGroups" in csharp_name:
            c |= c.CUSTOM_GROUPS
        if "Global" in csharp_name:
            c |= c.GLOBAL
        if "WitnessRules" in csharp_name:
            c |= c.WITNESS_RULES
        return c


class Signer(serialization.ISerializable, interfaces.IJson):
    """
    A class that specifies the rules of who can pass `CheckWitness()` verifications in a smart contract.
    """

    #: Maximum number of allowed_contracts or allowed_groups
    MAX_SUB_ITEMS = 16

    def __init__(
        self,
        account: types.UInt160,
        scope: WitnessScope = WitnessScope.CALLED_BY_ENTRY,
        allowed_contracts: Optional[Sequence[types.UInt160]] = None,
        allowed_groups: Optional[Sequence[cryptography.ECPoint]] = None,
        rules: Optional[Sequence[WitnessRule]] = None,
    ):
        #: The TX sender.
        self.account = account
        #: WitnessScope: The configured validation scope.
        self.scope = scope
        #: list[types.UInt160]: Whitelist of contract script hashes if used with `WitnessScope.CUSTOM_CONTRACTS`.
        self.allowed_contracts = allowed_contracts if allowed_contracts else []
        #: list[cryptography.ECPoint]: Whitelist of public keys if used with `WitnessScope.CUSTOM_GROUPS`.
        self.allowed_groups = allowed_groups if allowed_groups else []
        #: List of rules that must pass for the current execution context when used with `WitnessScope.WITNESS_RULES`.
        self.rules = rules if rules else []

    def __len__(self):
        contracts_size = 0
        if WitnessScope.CUSTOM_CONTRACTS in self.scope:
            contracts_size = utils.get_var_size(self.allowed_contracts)

        groups_size = 0
        if WitnessScope.CUSTOM_GROUPS in self.scope:
            groups_size = utils.get_var_size(self.allowed_groups)

        rules_size = 0
        if WitnessScope.WITNESS_RULES in self.scope:
            rules_size = utils.get_var_size(self.rules)

        return s.uint160 + s.uint8 + contracts_size + groups_size + rules_size

    def __eq__(self, other):
        if other is None:
            return False
        if type(self) != type(other):
            return False
        if self.account != other.account:
            return False
        return True

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_serializable(self.account)
        writer.write_uint8(self.scope)

        if WitnessScope.CUSTOM_CONTRACTS in self.scope:
            writer.write_serializable_list(self.allowed_contracts)

        if WitnessScope.CUSTOM_GROUPS in self.scope:
            writer.write_serializable_list(self.allowed_groups)

        if WitnessScope.WITNESS_RULES in self.scope:
            writer.write_serializable_list(self.rules)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        self.account = reader.read_serializable(types.UInt160)
        self.scope = WitnessScope(reader.read_uint8())

        if WitnessScope.GLOBAL in self.scope and self.scope != WitnessScope.GLOBAL:
            raise ValueError(
                "Deserialization error - invalid scope. GLOBAL scope not allowed with other scope types"
            )

        if WitnessScope.CUSTOM_CONTRACTS in self.scope:
            self.allowed_contracts = reader.read_serializable_list(
                types.UInt160, max=self.MAX_SUB_ITEMS
            )

        if WitnessScope.CUSTOM_GROUPS in self.scope:
            self.allowed_groups = reader.read_serializable_list(
                cryptography.ECPoint, max=self.MAX_SUB_ITEMS  # type: ignore
            )

        if WitnessScope.WITNESS_RULES in self.scope:
            self.rules = reader.read_serializable_list(
                WitnessRule, max=self.MAX_SUB_ITEMS
            )

    def get_all_rules(self) -> Iterator[WitnessRule]:
        """
        Return all witness rules.
        """
        if WitnessScope.GLOBAL in self.scope:
            yield WitnessRule(WitnessRuleAction.ALLOW, ConditionBool(True))
        else:
            if WitnessScope.CALLED_BY_ENTRY in self.scope:
                yield WitnessRule(WitnessRuleAction.ALLOW, ConditionCalledByEntry())

            if WitnessScope.CUSTOM_CONTRACTS in self.scope:
                for hash_ in self.allowed_contracts:
                    yield WitnessRule(
                        WitnessRuleAction.ALLOW, ConditionScriptHash(hash_)
                    )

            if WitnessScope.CUSTOM_GROUPS in self.scope:
                for group in self.allowed_groups:
                    yield WitnessRule(WitnessRuleAction.ALLOW, ConditionGroup(group))

            if WitnessScope.WITNESS_RULES in self.scope:
                for rule in self.rules:
                    yield rule

    def to_json(self) -> dict:
        """Convert object into JSON representation."""
        json: dict[str, Any] = {
            "account": "0x" + str(self.account),
            "scopes": self.scope.to_csharp_name(),
        }

        if WitnessScope.CUSTOM_CONTRACTS in self.scope:
            json.update(
                {
                    "allowedcontracts": list(
                        map(lambda a: "0x" + str(a), self.allowed_contracts)
                    )
                }
            )
        if WitnessScope.CUSTOM_GROUPS in self.scope:
            json.update(
                {"allowedgroups": list(map(lambda g: str(g), self.allowed_groups))}
            )
        if WitnessScope.WITNESS_RULES in self.scope:
            json.update({"rules": list(map(lambda r: r.to_json(), self.rules))})

        return json

    @classmethod
    def from_json(cls, json: dict):
        """Create object from JSON."""
        account = types.UInt160.from_string(json["account"][2:])
        scopes = WitnessScope.from_chsarp_name(json["scopes"])

        allowed_contracts = []
        if "allowedcontracts" in json:
            for contract in json["allowedcontracts"]:
                allowed_contracts.append(types.UInt160.from_string(contract[2:]))

        allowed_groups = []
        if "allowedgroups" in json:
            for group in json["allowedgroups"]:
                allowed_groups.append(
                    cryptography.ECPoint.deserialize_from_bytes(bytes.fromhex(group))
                )

        rules = []
        if "rules" in json:
            for rule in json["rules"]:
                rules.append(WitnessRule.from_json(rule))

        return cls(account, scopes, allowed_contracts, allowed_groups, rules)

    @classmethod
    def _serializable_init(cls):
        return cls(types.UInt160.zero())


class Witness(serialization.ISerializable, interfaces.IJson):
    """
    An executable verification script that validates a verifiable object like a transaction.
    """

    _MAX_INVOCATION_SCRIPT = 1024
    _MAX_VERIFICATION_SCRIPT = 1024

    def __init__(self, invocation_script: bytes, verification_script: bytes):
        #: A set of VM instructions to set up the stack for verification.
        self.invocation_script = invocation_script
        #: A set of VM instructions that does the actual verification.
        #: It is expected to set the result stack to a boolean True if validation passed.
        self.verification_script = verification_script
        self._script_hash = None

    def __len__(self):
        return utils.get_var_size(self.invocation_script) + utils.get_var_size(
            self.verification_script
        )

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_var_bytes(self.invocation_script)
        writer.write_var_bytes(self.verification_script)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        self.invocation_script = reader.read_var_bytes(max=self._MAX_INVOCATION_SCRIPT)
        self.verification_script = reader.read_var_bytes(
            max=self._MAX_VERIFICATION_SCRIPT
        )

    def script_hash(self) -> types.UInt160:
        """Get the script hash based on the verification script."""
        intermediate_data = hashlib.sha256(self.verification_script).digest()
        data = hashlib.new("ripemd160", intermediate_data).digest()
        return types.UInt160(data=data)

    def to_json(self) -> dict:
        """Convert object into JSON representation."""
        return {
            "invocation": base64.b64encode(self.invocation_script).decode(),
            "verification": base64.b64encode(self.verification_script).decode(),
        }

    @classmethod
    def from_json(cls, json: dict):
        """Create object from JSON"""
        return cls(
            base64.b64decode(json["invocation"]), base64.b64decode(json["verification"])
        )

    @classmethod
    def _serializable_init(cls):
        return cls(b"", b"")


class WitnessRuleAction(IntEnum):
    """
    Witness rule execution.
    """

    DENY = 0
    ALLOW = 1


class WitnessConditionType(IntEnum):
    """
    Type of valid witness conditions.
    """

    BOOLEAN = 0x0
    NOT = 0x01
    AND = 0x2
    OR = 0x03
    SCRIPT_HASH = 0x18
    GROUP = 0x19
    CALLED_BY_ENTRY = 0x20
    CALLED_BY_CONTRACT = 0x28
    CALLED_BY_GROUP = 0x29

    def to_csharp_string(self) -> str:
        if self == WitnessConditionType.SCRIPT_HASH:
            return "ScriptHash"
        elif self == WitnessConditionType.CALLED_BY_ENTRY:
            return "CalledByEntry"
        elif self == WitnessConditionType.CALLED_BY_CONTRACT:
            return "CalledByContract"
        elif self == WitnessConditionType.CALLED_BY_GROUP:
            return "CalledByGroup"
        else:
            return self.name.title()

    @classmethod
    def from_csharp_string(cls, name: str):
        match name:
            case "ScriptHash":
                return WitnessConditionType.SCRIPT_HASH
            case "CalledByEntry":
                return WitnessConditionType.CALLED_BY_ENTRY
            case "CalledByContract":
                return WitnessConditionType.CALLED_BY_CONTRACT
            case "CalledByGroup":
                return WitnessConditionType.CALLED_BY_GROUP
            case _:
                _name = name.upper()
                for k, v in cls.__members__.items():
                    if k == _name:
                        return v
                else:
                    raise ValueError(f"{name} cannot be converted to {cls.__name__}")


class WitnessCondition(serialization.ISerializable, interfaces.IJson):
    """
    Base class for conditions.
    """

    MAX_SUB_ITEMS = 16
    MAX_NESTING_DEPTH = 2

    _type = WitnessConditionType.BOOLEAN

    def __len__(self):
        return s.uint8  # 1 byte witness condition

    @property
    def type(self) -> WitnessConditionType:
        return self._type

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        writer.write_uint8(self.type.value)
        self._serialize_without_type(writer)

    @abc.abstractmethod
    def _serialize_without_type(self, writer: serialization.BinaryWriter) -> None:
        """ """

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        self._type = WitnessConditionType(reader.read_uint8())
        self._deserialize_without_type(reader, self.MAX_NESTING_DEPTH)

    @abc.abstractmethod
    def _deserialize_without_type(
        self, reader: serialization.BinaryReader, max_nesting_depth: int
    ) -> None:
        """Deserialize from a buffer without reading the `type` member."""

    @staticmethod
    def _deserialize_conditions(
        reader: serialization.BinaryReader, max_nesting_depth: int
    ) -> list[WitnessCondition]:
        conditions = []
        for _ in range(reader.read_var_int(WitnessCondition.MAX_SUB_ITEMS)):
            conditions.append(
                WitnessCondition._deserialize_from(reader, max_nesting_depth)
            )
        return conditions

    @staticmethod
    def _deserialize_from(
        reader: serialization.BinaryReader, max_nesting_depth: int
    ) -> WitnessCondition:
        condition_type = reader.read_uint8()

        def find_condition(condition) -> Optional[WitnessCondition]:
            for sub in condition.__subclasses__():
                child = sub._serializable_init()
                if child.type.value == condition_type:
                    child._deserialize_without_type(reader, max_nesting_depth)
                    return child
                if len(sub.__subclasses__()) > 0:
                    condition = find_condition(sub)
                    if condition is not None:
                        return condition
            return None

        condition = find_condition(WitnessCondition)
        if condition is None:
            raise ValueError(
                f"Deserialization error - unknown witness condition. Type: {hex(condition_type)}"
            )
        else:
            return condition

    def to_json(self) -> dict:
        """Convert object into JSON representation."""
        return {"type": self.type.to_csharp_string()}

    @classmethod
    def from_json(cls, json: dict):
        """Create object from JSON"""
        condition_type = WitnessConditionType.from_csharp_string(json["type"])
        for sub in WitnessCondition.__subclasses__():
            child = sub._serializable_init()  # type: ignore
            if child.type == condition_type:
                return child.from_json(json)


class ConditionAnd(WitnessCondition):
    """
    Match all conditions.
    """

    _type = WitnessConditionType.AND

    def __init__(self, expressions: list[WitnessCondition]):
        self.expressions = expressions

    def __len__(self):
        return super(ConditionAnd, self).__len__() + utils.get_var_size(
            self.expressions
        )

    def __eq__(self, other):
        if type(other) != type(self):
            return False
        return self.expressions == other.expressions

    def _serialize_without_type(self, writer: serialization.BinaryWriter) -> None:
        writer.write_serializable_list(self.expressions)

    def _deserialize_without_type(
        self, reader: serialization.BinaryReader, max_nesting_depth: int
    ) -> None:
        if max_nesting_depth <= 0:
            raise ValueError("Max nesting depth cannot be negative")
        self.expressions = WitnessCondition._deserialize_conditions(
            reader, max_nesting_depth
        )
        if len(self.expressions) == 0:
            raise ValueError("Cannot have 0 expressions")

    def to_json(self) -> dict:
        """Convert object into JSON representation."""
        json = super(ConditionAnd, self).to_json()
        json["expressions"] = list(map(lambda exp: exp.to_json(), self.expressions))
        return json

    @classmethod
    def from_json(cls, json: dict):
        return cls(
            list(
                map(lambda expr: WitnessCondition.from_json(expr), json["expressions"])
            )
        )

    @classmethod
    def _serializable_init(cls):
        return cls([])


class ConditionBool(WitnessCondition):
    """
    Fixed value. Can be used to emulate GLOBAL scope.
    """

    _type = WitnessConditionType.BOOLEAN

    def __init__(self, value: bool):
        self.value = value

    def __len__(self):
        return super(ConditionBool, self).__len__() + s.uint8

    def __eq__(self, other):
        if type(other) != type(self):
            return False
        return self.value == other.value

    def _serialize_without_type(self, writer: serialization.BinaryWriter) -> None:
        writer.write_bool(self.value)

    def _deserialize_without_type(
        self, reader: serialization.BinaryReader, max_nesting_depth: int
    ) -> None:
        self.value = reader.read_bool()

    def to_json(self) -> dict:
        """Convert object into JSON representation."""
        json = super(ConditionBool, self).to_json()
        json["expression"] = self.value
        return json

    @classmethod
    def from_json(cls, json: dict):
        return cls(json["expression"])

    @classmethod
    def _serializable_init(cls):
        return cls(False)


class ConditionNot(WitnessCondition):
    """
    Invert condition.
    """

    _type = WitnessConditionType.NOT

    def __init__(self, expression: WitnessCondition):
        self.expression = expression

    def __len__(self):
        return super(ConditionNot, self).__len__() + len(self.expression)

    def __eq__(self, other):
        if type(other) != type(self):
            return False
        return self.expression == other.expression

    def _serialize_without_type(self, writer: serialization.BinaryWriter) -> None:
        writer.write_serializable(self.expression)

    def _deserialize_without_type(
        self, reader: serialization.BinaryReader, max_nesting_depth: int
    ) -> None:
        if max_nesting_depth <= 0:
            raise ValueError("Max nesting depth cannot be negative")
        self.expression = WitnessCondition._deserialize_from(
            reader, max_nesting_depth - 1
        )

    def to_json(self) -> dict:
        """Convert object into JSON representation."""
        json = super(ConditionNot, self).to_json()
        json["expression"] = self.expression.to_json()
        return json

    @classmethod
    def from_json(cls, json: dict):
        return cls(WitnessCondition.from_json(json["expression"]))

    @classmethod
    def _serializable_init(cls):
        return cls(ConditionBool(False))


class ConditionOr(WitnessCondition):
    """
    Match any from a list of conditions.
    """

    _type = WitnessConditionType.OR

    def __init__(self, expressions: list[WitnessCondition]):
        self.expressions = expressions

    def __len__(self):
        return super(ConditionOr, self).__len__() + utils.get_var_size(self.expressions)

    def __eq__(self, other):
        if type(other) != type(self):
            return False
        return self.expressions == other.expressions

    def _serialize_without_type(self, writer: serialization.BinaryWriter) -> None:
        writer.write_serializable_list(self.expressions)

    def _deserialize_without_type(
        self, reader: serialization.BinaryReader, max_nesting_depth: int
    ) -> None:
        if max_nesting_depth <= 0:
            raise ValueError("Max nesting depth cannot be negative")
        self.expressions = WitnessCondition._deserialize_conditions(
            reader, max_nesting_depth
        )
        if len(self.expressions) == 0:
            raise ValueError("Cannot have 0 expressions")

    def to_json(self) -> dict:
        """Convert object into JSON representation."""
        json = super(ConditionOr, self).to_json()
        json["expressions"] = list(map(lambda exp: exp.to_json(), self.expressions))
        return json

    @classmethod
    def from_json(cls, json: dict):
        return cls(
            list(
                map(lambda expr: WitnessCondition.from_json(expr), json["expressions"])
            )
        )

    @classmethod
    def _serializable_init(cls):
        return cls([])


class ConditionCalledByContract(WitnessCondition):
    """
    Match hash against caller script hash.
    """

    _type = WitnessConditionType.CALLED_BY_CONTRACT

    def __init__(self, hash_: types.UInt160):
        self.hash_ = hash_

    def __len__(self):
        return super(ConditionCalledByContract, self).__len__() + s.uint160

    def __eq__(self, other):
        if type(other) != type(self):
            return False
        return self.hash_ == other.hash_

    def _serialize_without_type(self, writer: serialization.BinaryWriter) -> None:
        writer.write_serializable(self.hash_)

    def _deserialize_without_type(
        self, reader: serialization.BinaryReader, max_nesting_depth: int
    ) -> None:
        self.hash_ = reader.read_serializable(types.UInt160)

    def to_json(self) -> dict:
        """Convert object into JSON representation."""
        json = super(ConditionCalledByContract, self).to_json()
        json["hash"] = f"0x{self.hash_}"
        return json

    @classmethod
    def _serializable_init(cls):
        return cls(types.UInt160.zero())

    @classmethod
    def from_json(cls, json: dict):
        return cls(types.UInt160.from_string(json["hash"]))


class ConditionCalledByEntry(WitnessCondition):
    """
    Match hash against entry script hash.
    """

    _type = WitnessConditionType.CALLED_BY_ENTRY

    def __eq__(self, other):
        if type(self) == type(other):
            return True
        return False

    @classmethod
    def from_json(cls, json: dict):
        return cls()

    def _serialize_without_type(self, writer: serialization.BinaryWriter) -> None:
        pass

    def _deserialize_without_type(
        self, reader: serialization.BinaryReader, max_nesting_depth: int
    ) -> None:
        pass


class ConditionCalledByGroup(WitnessCondition):
    """
    Match key against the caller group.
    """

    _type = WitnessConditionType.CALLED_BY_GROUP

    def __init__(self, group: cryptography.ECPoint):
        self.group = group

    def __len__(self):
        return super(ConditionCalledByGroup, self).__len__() + len(self.group)

    def __eq__(self, other):
        if type(self) != type(other):
            return False

        return self.group == other.group

    def _serialize_without_type(self, writer: serialization.BinaryWriter) -> None:
        writer.write_serializable(self.group)

    def _deserialize_without_type(
        self, reader: serialization.BinaryReader, max_nesting_depth: int
    ) -> None:
        self.group = reader.read_serializable(cryptography.ECPoint)  # type: ignore

    def to_json(self) -> dict:
        """Convert object into JSON representation."""
        json = super(ConditionCalledByGroup, self).to_json()
        json["group"] = str(self.group)
        return json

    @classmethod
    def from_json(cls, json: dict):
        return cls(
            cryptography.ECPoint.deserialize_from_bytes(bytes.fromhex(json["group"]))
        )

    @classmethod
    def _serializable_init(cls):
        return cls(cryptography.ECPoint.deserialize_from_bytes(b"\x00"))


class ConditionGroup(ConditionCalledByGroup):
    """
    Match group against current script hash.
    """

    _type = WitnessConditionType.GROUP


class ConditionScriptHash(ConditionCalledByContract):
    """
    Match hash against another script hash.
    """

    _type = WitnessConditionType.SCRIPT_HASH


class WitnessRule(serialization.ISerializable, interfaces.IJson):
    """
    A firewall like rule with an action to deny or allow if the condition matches. Gives fine-grained control over
    where the signature of the witness is valid inside the system.
    """

    def __init__(self, action: WitnessRuleAction, condition: WitnessCondition):
        self.action = action
        self.condition = condition

    def __len__(self):
        return s.uint8 + len(self.condition)

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        writer.write_uint8(self.action.value)
        writer.write_serializable(self.condition)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        self.action = WitnessRuleAction(reader.read_uint8())
        self.condition = WitnessCondition._deserialize_from(
            reader, WitnessCondition.MAX_NESTING_DEPTH
        )

    def to_json(self) -> dict:
        """Convert object into JSON representation."""
        return {
            "action": self.action.name.title(),
            "condition": self.condition.to_json(),
        }

    @classmethod
    def from_json(cls, json: dict):
        """Create object from JSON."""
        action = WitnessRuleAction.DENY
        if json["action"] == "Allow":
            action = WitnessRuleAction.ALLOW
        condition = WitnessCondition.from_json(json["condition"])
        return cls(action, condition)

    @classmethod
    def _serializable_init(cls):
        return cls(WitnessRuleAction.DENY, ConditionBool(False))


class IVerifiable(serialization.ISerializable):
    """
    Verifiable interface.
    """

    def __init__(self, *args, **kwargs):
        super(IVerifiable, self).__init__(*args, **kwargs)
        self.witnesses: list[Witness] = []

    @abc.abstractmethod
    def serialize_unsigned(self, writer: serialization.BinaryWriter) -> None:
        """ """

    @abc.abstractmethod
    def deserialize_unsigned(self, reader: serialization.BinaryReader) -> None:
        """ """

    @abc.abstractmethod
    def get_script_hashes_for_verifying(self, snapshot) -> list[types.UInt160]:
        """
        Helper method to get the data used in verifying the object.
        """

    def get_hash_data(self, protocol_magic: int) -> bytes:
        """Get the unsigned data.

        Args:
            protocol_magic: network protocol number (NEO MainNet = 860833102, Testnet (T5) = 894710606, private net = ?)
        """
        with serialization.BinaryWriter() as writer:
            writer.write_uint32(protocol_magic)
            self.serialize_unsigned(writer)
            return writer.to_array()



"""
NEO RPC Node client and response classes.
"""

from __future__ import annotations

import socket

import aiohttp
import base64
import asyncio
import datetime
import time
from enum import Enum, IntEnum
from contextlib import suppress
from dataclasses import dataclass
from typing import (
    Optional,
    TypedDict,
    Any,
    Protocol,
    Iterator,
    Union,
    cast,
    Type,
    AsyncGenerator,
)
from collections.abc import Sequence
from neo3.core import types, cryptography, interfaces, serialization
from neo3.contracts import manifest, nef, contract, abi
from neo3.network.payloads import transaction, block, verification
from neo3.wallet import utils as walletutils


@dataclass
class BlockValidator:
    """
    Activate consensus member.
    """

    public_key: cryptography.ECPoint
    votes: int

    def __repr__(self):
        return f"{self.__class__.__name__}(public_key={self.public_key}, votes={self.votes}"


@dataclass
class Candidate(BlockValidator):
    """
    Candidate consensus member.
    """

    active: bool

    def __repr__(self):
        return f"{self.__class__.__name__}(public_key={self.public_key}, votes={self.votes}, active={self.active}"


@dataclass
class NextBlockValidatorsResponse:
    """
    Response to `getnextblockvalidators` RPC call.
    """

    validators: list[BlockValidator]

    @classmethod
    def from_json(cls, json: dict):
        nvr = cls([])
        for validator in json:
            pk = cryptography.ECPoint.deserialize_from_bytes(
                bytes.fromhex(validator["publickey"])
            )
            votes = int(validator["votes"])
            nvr.validators.append(BlockValidator(pk, votes))
        return nvr


@dataclass
class VersionProtocol:
    """
    Partial response of `getversion` RPC call.
    """

    address_version: int
    network: int
    validators_count: int
    ms_per_block: int
    max_traceable_blocks: int
    max_transactions_per_block: int
    max_valid_until_block_increment: int
    memorypool_max_transactions: int
    initial_gas_distribution: int
    hardforks: dict[str, int]


@dataclass
class GetVersionResponse:
    """
    Response to `getversion` RPC call.
    """

    tcp_port: int
    ws_port: Optional[int]
    nonce: int
    user_agent: str
    protocol: VersionProtocol
    rpc_session_enabled: bool
    rpc_max_iterator_results: int

    @classmethod
    def from_json(cls, json: dict):
        p = json["protocol"]

        hf = {}
        for pair in p["hardforks"]:
            hf.update({pair["name"]: pair["blockheight"]})

        vp = VersionProtocol(
            p["addressversion"],
            p["network"],
            p["validatorscount"],
            p["msperblock"],
            p["maxtraceableblocks"],
            p["maxtransactionsperblock"],
            p["maxvaliduntilblockincrement"],
            p["memorypoolmaxtransactions"],
            p["initialgasdistribution"],
            hf,
        )
        wsport = json.get("wsport", None)
        return cls(
            json["tcpport"],
            wsport,
            json["nonce"],
            json["useragent"],
            vp,
            json["rpc"]["sessionenabled"],
            json["rpc"]["maxiteratorresultitems"],
        )


@dataclass
class Peer:
    """
    P2P peer information.
    """

    address: str
    port: int


@dataclass
class GetPeersResponse:
    """
    Response to `getpeers` RPC call.
    """

    connected: list[Peer]
    bad: list[Peer]
    unconnected: list[Peer]

    @classmethod
    def from_json(cls, json: dict):
        c = cls([], [], [])
        for p in json["connected"]:
            c.connected.append(Peer(p["address"], p["port"]))
        for p in json["bad"]:
            c.bad.append(Peer(p["address"], p["port"]))
        for p in json["unconnected"]:
            c.unconnected.append(Peer(p["address"], p["port"]))
        return c


@dataclass
class Nep17Balance:
    """
    NEP-17 balance entry.
    """

    asset_hash: types.UInt160
    amount: int
    last_updated_block: int

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(asset_hash={self.asset_hash}, amount={self.amount}, "
            f"last_updated_block={self.last_updated_block})"
        )


@dataclass
class Nep17BalancesResponse:
    """
    Response to `getnep17balances` RPC call.
    """

    balances: list[Nep17Balance]
    address: str

    @classmethod
    def from_json(cls, json: dict):
        c = cls([], json["address"])
        for b in json["balance"]:
            h = types.UInt160.from_string(b["assethash"][2:])
            a = int(b["amount"])
            c.balances.append(Nep17Balance(h, a, b["lastupdatedblock"]))
        return c


@dataclass
class Nep17Transfer:
    """
    NEP-17 transfer record.
    """

    time: datetime.datetime
    asset_hash: types.UInt160
    transfer_address: str
    amount: int
    block_index: int
    transfer_notify_index: int
    tx_hash: types.UInt256

    @classmethod
    def from_json(cls, json: dict):
        time = datetime.datetime.fromtimestamp(
            json["timestamp"] / 1000, datetime.timezone.utc
        )
        hash_ = types.UInt160.from_string(json["assethash"][2:])
        transfer_addr = json["transferaddress"]
        amount = int(json["amount"])
        tx_hash = types.UInt256.from_string(json["txhash"][2:])
        return cls(
            time,
            hash_,
            transfer_addr,
            amount,
            json["blockindex"],
            json["transfernotifyindex"],
            tx_hash,
        )

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(time={self.time}, asset_hash={self.asset_hash}, "
            f"transfer_address={self.transfer_address}, amount={self.amount}, block_index={self.block_index}, "
            f"transfer_notify_index={self.transfer_notify_index}, tx_hash={self.tx_hash})"
        )


@dataclass
class Nep17TransfersResponse:
    """
    Response to `getnep17transfer` RPC call.
    """

    sent: list[Nep17Transfer]
    received: list[Nep17Transfer]
    address: str

    @classmethod
    def from_json(cls, json: dict):
        c = cls([], [], json["address"])
        for t in json["sent"]:
            c.sent.append(Nep17Transfer.from_json(t))
        for t in json["received"]:
            c.received.append(Nep17Transfer.from_json(t))
        return c


@dataclass
class MempoolResponse:
    """
    Response to `getrawmempool` RPC call.

    A verified transaction in the memory pool is a transaction which has had:
    - basic structural validation (e.g. max tx size)
    - signature validation
    - state validation
        - block validity expiration check
        - available balance vs network and system fees
        - etc
    """

    verified: list[types.UInt256]
    unverified: list[types.UInt256]

    @classmethod
    def from_json(cls, json: dict):
        c = cls([], [])
        for tx in json["verified"]:
            c.verified.append(types.UInt256.from_string(tx[2:]))
        for tx in json["unverified"]:
            c.unverified.append(types.UInt256.from_string(tx[2:]))
        return c


class StackItemType(Enum):
    """
    Virtual machine item types that can be found in the `stack` property of responses when executing a script or
     transactions.
    """

    ANY = "Any"
    ARRAY = "Array"
    BOOL = "Boolean"
    BUFFER = "Buffer"
    BYTE_STRING = "ByteString"
    INTEGER = "Integer"
    INTEROP_INTERFACE = "InteropInterface"
    MAP = "Map"
    POINTER = "Pointer"
    STRUCT = "Struct"


@dataclass
class StackItem:
    """
    Virtual machine stack item.
    """

    type: StackItemType
    value: Any

    def __repr__(self):
        return f"{self.__class__.__name__}(type={self.type.name}, value={self.value})"

    def as_bool(self) -> bool:
        """
        Unwrap as `bool`.

        Raises:
            ValueError: if internal item type does not match required.
        """
        if self.type != StackItemType.BOOL:
            raise ValueError(
                f"item is not of type '{StackItemType.BOOL}' but of type '{self.type}'"
            )
        return self.value

    def as_bytes(self) -> bytes:
        """
        Unwrap as `bytes`.

        Raises:
            ValueError: if internal item type does not match required.
        """
        if self.type not in (StackItemType.BYTE_STRING, StackItemType.BUFFER):
            raise ValueError(
                f"item is not of type '{StackItemType.BYTE_STRING}' or '{StackItemType.BUFFER} but of type '{self.type}'"
            )
        return self.value

    def as_str(self) -> str:
        """
        Unwrap as `str`.

        Raises:
            ValueError: if internal item type does not match required.
        """
        if self.type not in (StackItemType.BYTE_STRING, StackItemType.BUFFER):
            raise ValueError(
                f"item is not of type '{StackItemType.BYTE_STRING}' or '{StackItemType.BUFFER}' but of type '{self.type}'"
            )
        v = cast(bytes, self.value)
        return v.decode()

    def as_int(self) -> int:
        """
        Unwrap as `int`.

        Raises:
            ValueError: if internal item type does not match required.

        """
        if self.type != StackItemType.INTEGER:
            raise ValueError(
                f"item is not of type '{StackItemType.INTEGER}' but of type '{self.type}'"
            )
        v = cast(int, self.value)
        return v

    def as_uint160(self) -> types.UInt160:
        """
        Unwrap as `UInt160`.

        Raises:
            ValueError: if internal item type does not match required.
        """
        if self.type not in (StackItemType.BYTE_STRING, StackItemType.BUFFER):
            raise ValueError(
                f"item is not of type '{StackItemType.BYTE_STRING}' or '{StackItemType.BUFFER}' but of type '{self.type}'"
            )
        # we need to ensure the data is hex-escaped
        data = self.value
        with suppress(UnicodeDecodeError, ValueError):
            data = bytes.fromhex(data.decode())
        return types.UInt160(data)

    def as_uint256(self) -> types.UInt256:
        """
        Unwrap as `UInt256`.

        Raises:
            ValueError: if internal item type does not match required.
        """
        if self.type not in (StackItemType.BYTE_STRING, StackItemType.BUFFER):
            raise ValueError(
                f"item is not of type '{StackItemType.BYTE_STRING}' or '{StackItemType.BUFFER}' but of type '{self.type}'"
            )
        # we need to ensure the data is hex-escaped
        data = self.value
        with suppress(UnicodeDecodeError, ValueError):
            data = bytes.fromhex(data.decode())
        return types.UInt256(data)

    def as_address(self) -> str:
        """
        Unwrap as NEO3 address.
        """
        return walletutils.script_hash_to_address(self.as_uint160())

    def as_public_key(self) -> cryptography.ECPoint:
        """
        Unwrap as `ECPoint`.

        Raises:
            ValueError: if internal item type does not match required.
        """
        if self.type not in (StackItemType.BYTE_STRING, StackItemType.BUFFER):
            raise ValueError(
                f"item is not of type '{StackItemType.BYTE_STRING}' or '{StackItemType.BUFFER}' but of type '{self.type}'"
            )
        # we need to ensure the data is hex-escaped
        data = self.value
        with suppress(UnicodeDecodeError, ValueError):
            data = bytes.fromhex(data.decode())
        return cryptography.ECPoint.deserialize_from_bytes(data, validate=True)

    def as_list(self) -> list[StackItem]:
        """
        Unwrap as `list`.

        Raises:
            ValueError: if internal item type does not match required.
        """
        if self.type != StackItemType.ARRAY:
            raise ValueError(
                f"item is not of type '{StackItemType.ARRAY}' but of type '{self.type}'"
            )
        return cast(list, self.value)

    def as_dict(self) -> dict:
        """
        Unwrap as `dict`.

        Raises:
            ValueError: if internal item type does not match required.
        """
        if self.type != StackItemType.MAP:
            raise ValueError(
                f"item is not of type '{StackItemType.MAP}' but of type '{self.type}'"
            )
        m = cast(MapStackItem, self)
        return dict(m.items())

    def as_none(self) -> None:
        """
        Unwrap as `None`.

        Raises:
            ValueError: if internal item type does not match required.

        """
        if self.type != StackItemType.ANY:
            raise ValueError(
                f"item is not of type '{StackItemType.ANY}' but of type '{self.type}'"
            )
        if self.value is not None:
            raise ValueError(f"value is not None but of type '{type(self.value)}")
        return self.value


class MapStackItem(StackItem):
    def items(self) -> Iterator:
        for pair in self.value:  # type: tuple[StackItem, StackItem]
            yield pair[0].value, pair[1].value

    def keys(self) -> Iterator:
        for pair in self.value:  # type: tuple[StackItem, StackItem]
            yield pair[0].value

    def values(self) -> Iterator:
        for pair in self.value:  # type: tuple[StackItem, StackItem]
            yield pair[1].value

    def __getitem__(self, item: str):
        for pair in self.value:  # type: tuple[StackItem, StackItem]
            if pair[0].value == item:
                return pair[1].value
        else:
            raise KeyError

    def __iter__(self):
        for pair in self.value:  # type: tuple[StackItem, StackItem]
            yield pair[0].value


_Item = TypedDict("_Item", {"type": str, "value": Any})


@dataclass
class ExecutionResult:
    """
    Execution result data.
    """

    state: str
    gas_consumed: int
    exception: Optional[str]
    stack: list[StackItem]

    @staticmethod
    def _parse_stack_item(item: _Item) -> StackItem:
        try:
            type_ = StackItemType(item["type"])
        except ValueError:
            raise ValueError(f"Unknown stack item type: {item['type']}")

        if type_ in (StackItemType.ARRAY, StackItemType.STRUCT):
            list_ = list(
                map(
                    lambda element: ExecutionResult._parse_stack_item(element),
                    item["value"],
                )
            )
            return StackItem(type_, list_)
        elif type_ in (StackItemType.BOOL, StackItemType.POINTER):
            return StackItem(type_, item["value"])
        if type_ in (StackItemType.BUFFER, StackItemType.BYTE_STRING):
            return StackItem(type_, base64.b64decode(item["value"]))
        elif type_ == StackItemType.INTEGER:
            return StackItem(type_, int(item["value"]))
        elif type_ == StackItemType.MAP:
            map_ = []
            for stack_item in item["value"]:
                key = ExecutionResult._parse_stack_item(stack_item["key"])
                value = ExecutionResult._parse_stack_item(stack_item["value"])
                map_.append((key, value))
            return MapStackItem(type_, map_)
        elif type_ == StackItemType.ANY:
            return StackItem(type_, None)
        elif type_ == StackItemType.INTEROP_INTERFACE:
            if "iterator" not in item.keys():
                raise ValueError(
                    f"Interop stack item only supports iterators, could not find 'iterator' key"
                )
            values = list(
                map(
                    lambda element: ExecutionResult._parse_stack_item(element),
                    item["iterator"],  # type: ignore
                )
            )
            return StackItem(type_, values)
        assert False, "unreachable"

    @classmethod
    def from_json(cls, json: dict):
        gc = int(json["gasconsumed"])
        stack = list(
            map(lambda item: ExecutionResult._parse_stack_item(item), json["stack"])
        )
        return cls(json["state"], gc, json["exception"], stack)


@dataclass
class ExecutionResultResponse(ExecutionResult):
    """
    Response to `invokecontractverify`, `invokefunction` or `invokescript` RPC call.
    """

    script: bytes
    notifications: list[Notification]

    @classmethod
    def from_json(cls, json: dict):
        script = base64.b64decode(json["script"])
        gc = int(json["gasconsumed"])
        stack = list(
            map(lambda item: ExecutionResult._parse_stack_item(item), json["stack"])
        )
        notifications = []
        notifs = json.get("notifications", [])
        for n in notifs:
            notifications.append(Notification.from_json(n))

        return cls(
            json["state"],
            gc,
            json["exception"],
            stack,
            script,
            notifications=notifications,
        )


@dataclass
class Notification:
    """
    Smart contract notification entry.
    """

    contract: types.UInt160
    event_name: str
    state: StackItem

    @classmethod
    def from_json(cls, json: dict):
        c = types.UInt160.from_string(json["contract"][2:])
        e = json["eventname"]
        s = ExecutionResult._parse_stack_item(json["state"])
        return cls(c, e, s)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(contract={str(self.contract)}, "
            f"event_name={self.event_name}, state={self.state})"
        )


@dataclass
class ApplicationExecution(ExecutionResult):
    """
    Specialised `ExecutionResult` with additional notification and trigger information.
    """

    trigger: str
    notifications: list[Notification]

    @classmethod
    def from_json(cls, json: dict):
        gc = int(json["gasconsumed"])
        stack = list(
            map(lambda item: ExecutionResult._parse_stack_item(item), json["stack"])
        )
        state = json["vmstate"]
        ex = json.get("exception", None)
        notifications = []
        for n in json["notifications"]:
            notifications.append(Notification.from_json(n))

        return cls(
            trigger=json["trigger"],
            notifications=notifications,
            state=state,
            gas_consumed=gc,
            exception=ex,
            stack=stack,
        )


@dataclass
class TransactionApplicationLogResponse:
    """
    Data log for processing of a transaction on chain.
    """

    tx_hash: types.UInt256
    execution: ApplicationExecution

    @classmethod
    def from_json(cls, json: dict):
        tx_id = types.UInt256.from_string(json["txid"])
        execution = ApplicationExecution.from_json(json["executions"][0])
        return cls(tx_id, execution)

    def __repr__(self):
        return f"{self.__class__.__name__}(tx_hash={str(self.tx_hash)}, execution={self.execution})"


@dataclass
class BlockApplicationLogResponse:
    """
    Data log for processing of a block on chain.

    Does not include the data log for its transactions.
    """

    block_hash: types.UInt256
    executions: list[ApplicationExecution]

    @classmethod
    def from_json(cls, json: dict):
        block_hash = types.UInt256.from_string(json["blockhash"])
        executions = []
        for execution in json["executions"]:
            executions.append(ApplicationExecution.from_json(execution))
        return cls(block_hash, executions)

    def __repr__(self):
        return f"{self.__class__.__name__}(block_hash={str(self.block_hash)}, executions={self.executions})"


ContractParameter = Union[
    bool,
    int,
    str,
    bytes,
    bytearray,
    types.BigInteger,
    types.UInt160,
    types.UInt256,
    cryptography.ECPoint,
    "ContractParameterArray",
    "ContractParameterDict",
    Type[serialization.ISerializable_T],
]


class ContractParameterArray(Protocol):
    """"""

    def insert(self, index: int, value: ContractParameter) -> None: ...

    def __getitem__(self, i: int) -> ContractParameter: ...

    def __setitem__(self, i: int, o: ContractParameter) -> None: ...

    def __delitem__(self, i: int) -> None: ...


class ContractParameterDict(Protocol):
    """"""

    def __setitem__(self, k: ContractParameter, v: ContractParameter) -> None: ...

    def __delitem__(self, v: ContractParameter) -> None: ...

    def __getitem__(self, k: ContractParameter) -> ContractParameter: ...

    def __iter__(self) -> Iterator[ContractParameter]: ...


class _ContractParameter(interfaces.IJson):
    def __init__(self, obj: ContractParameter):
        self.value: ContractParameter = ""  # just to help mypy
        if isinstance(obj, bool):
            self.type = abi.ContractParameterType.BOOLEAN
            self.value = obj
        elif isinstance(obj, IntEnum):
            self.type = abi.ContractParameterType.INTEGER
            self.value = str(obj.value)
        elif isinstance(obj, (int, types.BigInteger)):
            self.type = abi.ContractParameterType.INTEGER
            self.value = str(obj)
        elif isinstance(obj, str):
            self.type = abi.ContractParameterType.STRING
            self.value = obj
        elif isinstance(obj, (bytes, bytearray)):
            self.type = abi.ContractParameterType.BYTEARRAY
            self.value = base64.b64encode(obj).decode()
        elif isinstance(obj, types.UInt160):
            self.type = abi.ContractParameterType.HASH160
            self.value = f"0x{obj}"
        elif isinstance(obj, types.UInt256):
            self.type = abi.ContractParameterType.HASH256
            self.value = f"0x{obj}"
        elif isinstance(obj, cryptography.ECPoint):
            self.type = abi.ContractParameterType.PUBLICKEY
            self.value = obj.to_array().hex()
        elif isinstance(obj, (list, tuple)):
            self.type = abi.ContractParameterType.ARRAY
            self.value = list(map(lambda element: _ContractParameter(element), obj))
        elif isinstance(obj, dict):
            self.type = abi.ContractParameterType.MAP
            pairs: list[dict] = []
            for k, v in obj.items():
                pairs.append(
                    {"key": _ContractParameter(k), "value": _ContractParameter(v)}
                )
            # It seems like mypy can't follow that ContractParameter is also
            # a list[dict[ContractParameter, ContractParameter]]
            self.value = pairs  # type: ignore
        elif isinstance(obj, serialization.ISerializable):
            self.type = abi.ContractParameterType.BYTEARRAY
            self.value = base64.b64encode(obj.to_array()).decode()
        else:
            raise ValueError(f"Unsupported type {type(obj)}")

    @classmethod
    def from_json(cls, json: dict):
        """Not supported."""
        raise NotImplementedError

    def to_json(self) -> dict:
        """
        Convert object into JSON representation.
        """
        return {"type": self.type.PascalCase(), "value": self.value}


class RPCClient:
    """
    RPC Client base.
    """

    def __init__(self, url: str, timeout: float = 3.0):
        """
        Args:
            url: host + port.
            timeout: total time in seconds a request may take.
        """
        self.url = url
        self.timeout = timeout
        self.session = aiohttp.ClientSession(
            timeout=aiohttp.ClientTimeout(total=self.timeout),
            connector=aiohttp.TCPConnector(family=socket.AF_INET),
        )

    async def _post(self, json: dict):
        """
        Create a POST request with JSON to `self.url` with `self.timeout`.

        Raises:
            asyncio.exceptions.TimeoutError
        """
        async with self.session.post(self.url, json=json) as request:
            return await request.json(content_type=None)

    async def close(self):
        """
        Close the client session.
        """
        await self.session.close()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if not self.session.closed:
            await self.session.close()


class JsonRpcError(Exception):
    def __init__(self, code: int, message: str, data: Optional[str] = None):
        self.code = code
        self.message = message
        self.data = "" if data is None else data

    def __str__(self):
        if len(self.data) > 0:
            return f"code={self.code}, message={self.message}, data={self.data}"
        else:
            return f"code={self.code}, message={self.message}"


class JsonRpcTimeoutError(JsonRpcError):
    def __init__(self, message: Optional[str] = None):
        self.message = "Operation timed out" if message is None else message

    def __str__(self):
        return self.message


class NeoRpcClient(RPCClient):
    """
    Specialised RPC client for NEO's Node RPC API.
    """

    def __init__(self, host: str, **kwargs):
        super(NeoRpcClient, self).__init__(host, **kwargs)

    async def _do_post(
        self,
        method: str,
        params: Optional[list] = None,
        id: int = 0,
        jsonrpc_version: str = "2.0",
    ):
        params = params if params else []
        json = {
            "jsonrpc": jsonrpc_version,
            "id": id,
            "method": method,
            "params": params,
        }
        response = await super(NeoRpcClient, self)._post(json)
        if "error" in response:
            raise JsonRpcError(**response["error"])
        return response["result"]

    async def calculate_network_fee(self, tx: bytes | transaction.Transaction) -> int:
        """
        Obtain the cost of verifying the transaction and including it in a block (a.k.a network fee).
        """
        if isinstance(tx, transaction.Transaction):
            tx = tx.to_array()
        params = [base64.b64encode(tx).decode()]
        result = await self._do_post("calculatenetworkfee", params)
        return int(result["networkfee"])

    async def find_states(
        self, contract_hash: types.UInt160 | str, prefix: Optional[bytes] = None
    ) -> AsyncGenerator[tuple[bytes, bytes], None]:
        """
        Fetch the smart contract storage state.

        Args:
            contract_hash: the hash of the smart contract to call.
            prefix: storage prefix to search for. If omitted will return all storage

        Returns:
            a storage key/value pair

        Examples:
            # prints all deployed
            prefix_contract_hash = b"\x0c"
            async with api.NeoRpcClient("https://testnet1.neo.coz.io:443") as client:
                async for k, v in client.find_states(CONTRACT_HASHES.MANAGEMENT, prefix_contract_hash):
                    print(k, v)

        """
        if isinstance(contract_hash, str):
            contract_hash = types.UInt160.from_string(contract_hash)
        contract_hash = f"0x{contract_hash}"

        if prefix is None:
            prefix = b""
        _prefix = base64.b64encode(prefix).decode()
        start = 0
        while True:
            response = await self._do_post(
                "findstorage", [contract_hash, _prefix, start]
            )
            if response["results"] is not None:
                for pair in response["results"]:
                    key = base64.b64decode(pair["key"])
                    value = base64.b64decode(pair["value"])
                    yield key, value
            if not response["truncated"]:
                break
            start = response["next"]

    async def get_application_log_transaction(
        self, tx_hash: types.UInt256 | str
    ) -> TransactionApplicationLogResponse:
        """
        Fetch the smart contract event logs for a given transaction.

        Commonly used to verify that a transaction sent via `send_transaction()` was executed succesfully on chain.

        Args:
            tx_hash: the hash of the transaction to query for.
        """
        if isinstance(tx_hash, types.UInt256):
            tx_hash = f"0x{str(tx_hash)}"
        result = await self._do_post("getapplicationlog", [tx_hash])
        return TransactionApplicationLogResponse.from_json(result)

    async def get_application_log_block(
        self, block_hash: types.UInt256 | str
    ) -> BlockApplicationLogResponse:
        """
        Fetch the system event logs for a given block.

        Args:
            block_hash: the hash of the block to query for.
        """
        if isinstance(block_hash, types.UInt256):
            block_hash = f"0x{str(block_hash)}"
        result = await self._do_post("getapplicationlog", [block_hash])
        return BlockApplicationLogResponse.from_json(result)

    async def get_best_block_hash(self) -> types.UInt256:
        """
        Fetch the hash of the highest block in the chain.
        """
        response = await self._do_post("getbestblockhash")
        return types.UInt256.from_string(response[2:])

    async def get_block(self, index_or_hash: int | types.UInt256) -> block.Block:
        """
        Fetch the block by its index or block hash.
        """
        params: list[int | str] = []
        if isinstance(index_or_hash, types.UInt256):
            params.append(f"0x{index_or_hash}")
        else:
            params.append(index_or_hash)
        response = await self._do_post("getblock", params)
        return block.Block.deserialize_from_bytes(base64.b64decode(response))

    async def get_block_count(self) -> int:
        """
        Fetch the current height of the blockchain.
        """
        return await self._do_post("getblockcount")

    async def get_block_hash(self, index: int) -> types.UInt256:
        """
        Fetch the block hash by the block's index.
        """
        response = await self._do_post("getblockhash", [index])
        return types.UInt256.from_string(response[2:])

    async def get_block_header(
        self, index_or_hash: int | types.UInt256
    ) -> block.Header:
        """
        Fetch the block header by its index or block hash.
        """
        if isinstance(index_or_hash, types.UInt256):
            params = [f"0x{index_or_hash}"]
        else:
            params = [str(index_or_hash)]
        response = await self._do_post("getblockheader", params)
        return block.Header.deserialize_from_bytes(base64.b64decode(response))

    async def get_candidates(self) -> Sequence[Candidate]:
        """
        Fetch list of consensus candidates.
        """
        response = await self._do_post("getcandidates")
        candidates = []
        for candidate in response:
            pk = cryptography.ECPoint.deserialize_from_bytes(
                bytes.fromhex(candidate["publickey"])
            )
            votes = int(candidate["votes"])
            candidates.append(Candidate(pk, votes, candidate["active"]))
        return candidates

    async def get_committee(self) -> tuple[cryptography.ECPoint, ...]:
        """
        Fetch the public keys of the current NEO committee.
        """
        response = await self._do_post("getcommittee")
        return tuple(
            map(
                lambda pk: cryptography.ECPoint.deserialize_from_bytes(
                    bytes.fromhex(pk)
                ),
                response,
            )
        )

    async def get_connection_count(self) -> int:
        """
        Fetch the number of peers connected to the node.
        """
        return await self._do_post("getconnectioncount")

    async def get_contract_state(
        self, contract_hash_or_name: types.UInt160 | str
    ) -> contract.ContractState:
        """
        Fetch smart contract state information.

        Note:
            Only native contracts can be queried by their name. Name is case-insensitive.
        """
        if isinstance(contract_hash_or_name, types.UInt160):
            params = [f"0x{str(contract_hash_or_name)}"]
        else:
            params = [contract_hash_or_name]
        result = await self._do_post("getcontractstate", params)

        h = types.UInt160.from_string(result["hash"][2:])
        nef_ = nef.NEF.from_json(result["nef"])
        manifest_ = manifest.ContractManifest.from_json(result["manifest"])
        return contract.ContractState(
            result["id"], nef_, manifest_, result["updatecounter"], h
        )

    async def get_nep17_balances(self, address: str) -> Nep17BalancesResponse:
        """
        Fetch the balance of all NEP17 assets for the specified address.
        """
        result = await self._do_post("getnep17balances", [address])
        return Nep17BalancesResponse.from_json(result)

    async def get_nep17_transfers(
        self,
        address: str,
        start_time: Optional[datetime.datetime] = None,
        end_time: Optional[datetime.datetime] = None,
    ) -> Nep17TransfersResponse:
        """
        Fetch NEP17 transfers for a given address. Defaults to the last 7 days on the server side.

        Args:
            address: account to get transfer for.
            start_time: if given the start of the requested range. Must be in UTC and time aware not naïve.
            end_time: if given the end of the requested range.

        Example:
            # Fetch transfers of the last 14 days
            start = datetime.datetime.now(datetime.timezone.utc) - datetime.timedelta(days=14)
            await get_nep17_transfers(<your address>, start)
        """
        params = [address]
        if start_time is not None:
            if start_time.tzinfo is None:
                raise ValueError(
                    "start_time is a naïve datetime object which can cause incorrect results. Make it "
                    "time aware by adding tzinfo. For more information see: "
                    "https://docs.python.org/3/library/datetime.html#datetime.datetime.tzinfo"
                )
            # C# server side expects timestamp in milliseconds instead of seconds
            t = int(start_time.timestamp() * 1000)
            params.append(str(t))

        if end_time is not None:
            if end_time.tzinfo is None:
                raise ValueError(
                    "end_time is a naïve object which can cause incorrect results. Make it time aware by "
                    "adding tzinfo. For more information see: "
                    "https://docs.python.org/3/library/datetime.html#datetime.datetime.tzinfo"
                )
            t = int(end_time.timestamp() * 1000)
            params.append(str(t))

        result = await self._do_post("getnep17transfers", params)
        return Nep17TransfersResponse.from_json(result)

    async def get_raw_mempool(self) -> MempoolResponse:
        """
        Fetch the transaction hashes currently in the memory pool waiting to be added to the next produced block.
        """
        result = await self._do_post("getrawmempool", [True])
        return MempoolResponse.from_json(result)

    async def get_next_blockvalidators(self) -> NextBlockValidatorsResponse:
        """
        Fetch the list of next block validators.
        """
        result = await self._do_post("getnextblockvalidators")
        return NextBlockValidatorsResponse.from_json(result)

    async def get_peers(self) -> GetPeersResponse:
        """
        Fetch peer information.
        """
        result = await self._do_post("getpeers")
        return GetPeersResponse.from_json(result)

    async def get_storage(self, script_hash: types.UInt160, key: bytes) -> bytes:
        """
        Fetch a value from a smart contracts storage by its key.

        Args:
            script_hash: contract script hash.
            key: the storage key to fetch the data for.

        Example:
            # fetch the fee per byte from the Policy native contract
            key_fee_per_byte = b'\x0a'
            await client.get_storage(contracts.PolicyContract().hash, key_fee_per_byte)
        """
        hash_ = f"0x{script_hash}"
        key_encoded = base64.b64encode(key).decode()
        result = await self._do_post("getstorage", params=[hash_, key_encoded])
        return base64.b64decode(result)

    async def get_transaction(
        self, tx_hash: types.UInt256 | str
    ) -> transaction.Transaction:
        """
        Fetch a transaction by its hash.
        """
        if isinstance(tx_hash, str):
            tx_hash = types.UInt256.from_string(tx_hash)
        result = await self._do_post("getrawtransaction", [f"0x{tx_hash}"])
        return transaction.Transaction.deserialize_from_bytes(base64.b64decode(result))

    async def get_transaction_height(self, tx_hash: types.UInt256 | str) -> int:
        """
        Fetch the height of the block the transaction is included in.
        """
        if isinstance(tx_hash, str):
            tx_hash = types.UInt256.from_string(tx_hash)
        return await self._do_post("gettransactionheight", [f"0x{tx_hash}"])

    async def get_unclaimed_gas(self, address: str) -> int:
        """
        Fetch the amount of unclaimed gas for the given address.

        Args:
            address: a NEO address.
        """
        result = await self._do_post("getunclaimedgas", [address])
        return int(result["unclaimed"])

    async def get_version(self) -> GetVersionResponse:
        """
        Fetch the node client version, network protocol properties and network ports.
        """
        return GetVersionResponse.from_json(await self._do_post("getversion"))

    async def invoke_contract_verify(
        self,
        contract_hash: types.UInt160 | str,
        function_params: Optional[Sequence[ContractParameter]] = None,
        signers: Optional[Sequence[verification.Signer]] = None,
    ) -> ExecutionResultResponse:
        """
        Invoke the `verify` method on the contract.

        Note:
            Calling smart contracts through this function does not alter the blockchain state.
            The smart contract will be called using the Verification trigger (unlike the `invoke_function` method
            which uses the Application trigger).

        Args:
            contract_hash: the hash of the smart contract to call.
            function_params: the arguments required by the smart contract function.
            signers: additional signers (e.g. for checkwitness passing).
        """
        if isinstance(contract_hash, str):
            contract_hash = types.UInt160.from_string(contract_hash)
        contract_hash = f"0x{contract_hash}"

        function_params = [] if function_params is None else function_params
        function_params = list(
            map(lambda fp: _ContractParameter(fp).to_json(), function_params)
        )

        signers = [] if signers is None else signers
        signers = list(map(lambda s: s.to_json(), signers))  # type: ignore

        params = [contract_hash, function_params, signers]
        result = await self._do_post("invokecontractverify", params)
        return ExecutionResultResponse.from_json(result)

    async def invoke_function(
        self,
        contract_hash: types.UInt160 | str,
        name: str,
        function_params: Optional[Sequence[ContractParameter]] = None,
        signers: Optional[Sequence[verification.Signer]] = None,
    ) -> ExecutionResultResponse:
        """
        Invoke a smart contract function.

        Note:
            Calling smart contracts through this function does not alter the blockchain state.
            To alter the blockchain state use the `send_transaction` method instead.

        Args:
            contract_hash: the hash of the smart contract to call.
            name: the name of the function to call on the smart contract.
            function_params: the arguments required by the smart contract function.
            signers: additional signers (e.g. for checkwitness passing).

        Example:
            # check if an account is blocked using the Policy native contract
            policy_contract = "cc5e4edd9f5f8dba8bb65734541df7a1c081c67b"
            account_to_check = types.UInt160.from_string("86df72a6b4ab5335d506294f9ce993722253b6e2")
            signer_account = types.UInt160.from_string("f621168b1fce3a89c33a5f6bcf7e774b4657031c")
            signer = verification.Signer(signer_account, payloads.WitnessScope.CALLED_BY_ENTRY)
            await client.invoke_function(contract_hash=policy_contract, name="isBlocked",
                                         function_params=[account_to_check], signers=[signer])
        """
        if isinstance(contract_hash, str):
            contract_hash = types.UInt160.from_string(contract_hash)
        contract_hash = f"0x{contract_hash}"

        function_params = [] if function_params is None else function_params
        function_params = list(
            map(lambda fp: _ContractParameter(fp).to_json(), function_params)
        )

        signers = [] if signers is None else signers
        signers = list(map(lambda s: s.to_json(), signers))  # type: ignore

        params = [contract_hash, name, function_params, signers]
        result = await self._do_post("invokefunction", params)
        return ExecutionResultResponse.from_json(result)

    async def invoke_script(
        self, script: bytes, signers: Optional[Sequence[verification.Signer]] = None
    ) -> ExecutionResultResponse:
        """
        Execute a script in the virtual machine.

        Note:
            Executing VM scripts through this function does not alter the blockchain state.

        Args:
            script: an array of VM opcodes.
            signers: additional signers (e.g. for checkwitness passing).

        Returns:
            The results of executing the script in the VM.
        """
        signers = [] if signers is None else signers
        signers = list(map(lambda s: s.to_json(), signers))  # type: ignore

        params = [base64.b64encode(script).decode(), signers]
        result = await self._do_post("invokescript", params)
        return ExecutionResultResponse.from_json(result)

    async def send_transaction(
        self, tx: transaction.Transaction | bytes
    ) -> types.UInt256:
        """
        Broadcast a transaction to the network.

        Note:
            uses the `sendrawtransaction` RPC method internally.

        Args:
            tx: either a Transaction object or a serialized Transaction. Must be signed.

        Returns:
            a transaction hash if successful.
        """
        if isinstance(tx, transaction.Transaction):
            tx = tx.to_array()
        result = await self._do_post(
            "sendrawtransaction", [base64.b64encode(tx).decode()]
        )
        return types.UInt256.from_string(result["hash"][2:])

    async def send_block(self, block_: block.Block | bytes) -> types.UInt256:
        """
        Broadcast a transaction to the network.

        Args:
            block_: either a Block object or a serialized Block.

        Returns:
            a block hash if successful.
        """
        if isinstance(block_, block.Block):
            block_ = block_.to_array()
        result = await self._do_post("submitblock", [base64.b64encode(block_).decode()])
        return types.UInt256.from_string(result["hash"][2:])

    async def validate_address(self, address: str) -> bool:
        """
        Verify if the given address is valid for the network the node is running on.

        Args:
            address: a NEO address.
        """
        result = await self._do_post("validateaddress", [address])
        return result["isvalid"]

    async def print_contract_methods(
        self, contract_hash_or_name: types.UInt160 | str
    ) -> None:
        """
        Helper to fetch all public methods of a smart contract, print their signatures in Python syntax as
        to help determine the right native argument types.

        Note:
            Only native contracts can be queried by their name. Name is case-insensitive.
        """
        state = await self.get_contract_state(contract_hash_or_name)

        print(f"Contract: {state.manifest.name}")
        print((10 + len(state.manifest.name)) * "-")
        for method in state.manifest.abi.methods:
            params = map(
                lambda p: f", {p.name}: {self._contract_param_to_native(p.type)}",
                method.parameters,
            )
            params = "".join(params)  # type: ignore

            # return types are not included because ABI types like ARRAY cannot be properly translated e.g. the
            # following functions both have ARRAY as return type in the ABI but their actual response is very different
            # 1. NeoToken.GetNextBlockValidators returns a list of serialized ECPoints
            # 2. ManagementContract.getContract a serialized ContractState (not even a list)
            print(f"def {method.name}(self{params})")
        print(" ")
        print(
            "ContractParam = Union[bool, int, str, bytes, UInt160, UInt256, ECPoint, list[ContractParam], "
            "dict[ContractParam, ContractParam]"
        )

    @staticmethod
    def _contract_param_to_native(p: abi.ContractParameterType) -> str:
        if p == abi.ContractParameterType.BOOLEAN:
            return "bool"
        elif p == abi.ContractParameterType.INTEGER:
            return "int"
        elif p == abi.ContractParameterType.STRING:
            return "str"
        elif p == abi.ContractParameterType.BYTEARRAY:
            return "bytes"
        elif p == abi.ContractParameterType.HASH160:
            return "UInt160"
        elif p == abi.ContractParameterType.HASH256:
            return "UInt256"
        elif p == abi.ContractParameterType.PUBLICKEY:
            return "ECPoint"
        elif p == abi.ContractParameterType.ARRAY:
            return "list"
        elif p == abi.ContractParameterType.MAP:
            return "dict"
        elif p == abi.ContractParameterType.VOID:
            return "None"
        elif p == abi.ContractParameterType.ANY:
            return "ContractParam"
        else:
            return f"Unknown({str(p)}"

    async def get_transaction_receipt(self, tx_hash: types.UInt256) -> Receipt:
        """
        Fetch a transaction receipt.

        Args:
            tx_hash: unique identifier of the transaction to fetch receipt for.
        """
        log = await self.get_application_log_transaction(tx_hash)
        included_in = await self.get_transaction_height(tx_hash)
        confirmations = await self.get_block_count() - included_in

        return Receipt(tx_hash, included_in, confirmations, log.execution)

    async def wait_for_transaction_receipt(
        self, tx_hash: types.UInt256, timeout: float = 20, retry_delay: float = 5
    ) -> Receipt:
        """
        Try to fetch a transaction.

        Args:
            tx_hash: unique identifier of the transaction to fetch the receipt for.
            timeout: maximum time to wait to find the transaction on chain in seconds.
            retry_delay: interval between querying the chain for the transaction in seconds.

        Raises:
            JsonRpcTimeoutError: if timeout threshold is exceeded
            JsonRpcError: for other errors that might occur.
        """
        start = time.time()
        while time.time() - start < timeout:
            try:
                return await self.get_transaction_receipt(tx_hash)
            except JsonRpcError as e:
                if (
                    "Unknown transaction" in e.message
                    or "Unknown script container" in e.message
                ):
                    await asyncio.sleep(retry_delay)
                else:
                    raise e
        else:
            raise JsonRpcTimeoutError(
                f"Could not find receipt for {tx_hash} within specified timeout of {timeout} seconds"
            )


@dataclass
class Receipt:
    """
    Transaction receipt containing data regarding chain state and events happening as a result of executing the
     transaction.
    """

    tx_hash: types.UInt256
    included_in_block: int
    confirmations: int
    execution: ApplicationExecution



import unittest
from neo3 import vm
from neo3.core import types, cryptography
from neo3.contracts import callflags


class ScriptBuilderTestCase(unittest.TestCase):
    def shortDescription(self):
        # disable docstring printing in test runner
        return None

    def test_emit(self):
        # test emit opcode without operand
        sb = vm.ScriptBuilder()
        sb.emit(vm.OpCode.DUP)
        self.assertEqual(vm.OpCode.DUP, sb.to_array())

        # test emit opcode withop operand
        sb = vm.ScriptBuilder()
        sb.emit(vm.OpCode.PUSHDATA1, b"\x01")
        self.assertEqual(vm.OpCode.PUSHDATA1 + b"\x01", sb.to_array())

    def test_emit_push_simple(self):
        sb = vm.ScriptBuilder()
        sb.emit_push(None)
        sb.emit_push(True)
        sb.emit_push(False)
        expected = vm.OpCode.PUSHNULL + vm.OpCode.PUSHT + vm.OpCode.PUSHF
        self.assertEqual(expected, sb.to_array())

    def test_emit_push_strings(self):
        sb = vm.ScriptBuilder()
        sb.emit_push("hello")
        # captured from C#
        expected = "0c0568656c6c6f"
        self.assertEqual(expected, sb.to_array().hex())

        sb = vm.ScriptBuilder()
        sb.emit_push("привет")
        expected = "0c0cd0bfd180d0b8d0b2d0b5d182"
        self.assertEqual(expected, sb.to_array().hex())

        sb = vm.ScriptBuilder()
        sb.emit_push("")
        expected = "0c00"
        self.assertEqual(expected, sb.to_array().hex())

    def test_emit_push_uint(self):
        sb = vm.ScriptBuilder()
        sb.emit_push(
            types.UInt160.from_string("0x6f1837723768f27a6f6a14452977e3e0e264f2cc")
        )
        # captured from C#
        expected = "0c14ccf264e2e0e3772945146a6f7af268377237186f"
        self.assertEqual(expected, sb.to_array().hex())

    def test_emit_push_public_key(self):
        key_raw = bytes.fromhex(
            "03b209fd4f53a7170ea4444e0cb0a6bb6a53c2bd016926989cf85f9b0fba17a70c"
        )
        pub_key = cryptography.ECPoint.deserialize_from_bytes(key_raw)
        sb = vm.ScriptBuilder()
        sb.emit_push(pub_key)
        # captured from C#
        expected = (
            "0c2103b209fd4f53a7170ea4444e0cb0a6bb6a53c2bd016926989cf85f9b0fba17a70c"
        )
        self.assertEqual(expected, sb.to_array().hex())

    def test_emit_push_enum(self):
        sb = vm.ScriptBuilder()
        sb.emit_push(vm.OpCode.DUP)
        # captured from C#
        expected = "004a"
        self.assertEqual(expected, sb.to_array().hex())

    def test_emit_push_small_numbers(self):
        sb = vm.ScriptBuilder()
        for i in range(16 + 1):
            sb.emit_push(i)
        # captured from C#
        expected = "101112131415161718191a1b1c1d1e1f20"
        self.assertEqual(expected, sb.to_array().hex())

        # check with BigInteger type
        sb = vm.ScriptBuilder()
        sb.emit_push(types.BigInteger.one())
        self.assertEqual(b"\x11", sb.to_array())

    def test_emit_push_bigger_numbers(self):
        sb = vm.ScriptBuilder()
        sb.emit_push(0x1F)  # PUSHINT8
        sb.emit_push(0xFF)  # PUSHINT16
        sb.emit_push(0xFFFF)  # PUSHINT32
        sb.emit_push(0xFFFFFFFF)  # PUSHINT64
        sb.emit_push(0xFFFFFFFF_FFFFFFFF)  # PUSHINT128
        sb.emit_push(types.BigInteger(b"\x01" * 17))  # PUSHINT256

        # captured from C#
        expected = "001f01ff0002ffff000003ffffffff0000000004ffffffffffffffff0000000000000000050101010101010101010101010101010101000000000000000000000000000000"
        self.assertEqual(expected, sb.to_array().hex())

        with self.assertRaises(ValueError) as context:
            sb.emit_push(types.BigInteger(b"\x01" * 33))
        self.assertEqual(
            "Input number exceeds maximum data size of 32 bytes", str(context.exception)
        )

    def test_emit_push_bytes(self):
        sb = vm.ScriptBuilder()
        sb.emit_push(b"\x01")
        expected = "0c0101"
        self.assertEqual(expected, sb.to_array().hex())
        self.assertEqual(vm.OpCode.PUSHDATA1 + b"\x01\x01", sb.to_array())

        data = b"\x01" * 0x100
        sb = vm.ScriptBuilder()
        sb.emit_push(data)
        expected = "0d000101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101"
        self.assertEqual(expected, sb.to_array().hex())

        data = b"\x01" * 0x10000
        sb = vm.ScriptBuilder()
        sb.emit_push(data)
        expected_header = bytes.fromhex("0e00000100")
        self.assertEqual(expected_header + data, sb.to_array())

        data = b"\x01" * 0x10001
        sb = vm.ScriptBuilder()
        sb.emit_push(data)
        expected_header = bytes.fromhex("0e01000100")
        self.assertEqual(expected_header + data, sb.to_array())

        # too slow
        # data = b'\x01' * (0xFFFFFFFF + 1)
        # sb = vm.ScriptBuilder()
        # with self.assertRaises(ValueError) as context:
        #     sb.emit_push(data)
        # self.assertIn("Value is too long", str(context.exception))

    def test_emit_push_dict(self):
        data = {"a": 123, "b": 456}

        sb = vm.ScriptBuilder()
        sb.emit_push(data)
        expected = "01c8010c0162007b0c016112be"
        self.assertEqual(expected, sb.to_array().hex())

        data = {b"\x01": 1, b"\x02": 2}
        sb = vm.ScriptBuilder()
        sb.emit_push(data)
        expected = "120c0102110c010112be"
        self.assertEqual(expected, sb.to_array().hex())

        # test invalid key type
        sb = vm.ScriptBuilder()
        with self.assertRaises(ValueError) as context:
            sb.emit_push({1.0: "abc"})
        self.assertEqual(
            "Unsupported key type <class 'float'>. Supported types by the VM are bool, int, str, bytes or ISerializable",
            str(context.exception),
        )

    def test_emit_push_list(self):
        data = ["a", 123, "b", 456]

        sb = vm.ScriptBuilder()
        sb.emit_push(data)
        expected = "01c8010c0162007b0c016114c0"
        self.assertEqual(expected, sb.to_array().hex())

    def test_emit_push_unsupported(self):
        class Unsupported:
            pass

        sb = vm.ScriptBuilder()
        with self.assertRaises(ValueError) as context:
            sb.emit_push(Unsupported())
        self.assertIn("Unsupported value type", str(context.exception))

    def test_emit_jump(self):
        sb = vm.ScriptBuilder()
        sb.emit_jump(vm.OpCode.JMP, 127)
        sb.emit_jump(vm.OpCode.JMP, 128)

        # captured from C#
        expected = "227f2380000000"
        self.assertEqual(expected, sb.to_array().hex())

        sb = vm.ScriptBuilder()
        with self.assertRaises(ValueError) as context:
            sb.emit_jump(vm.OpCode.DUP, 1)
        self.assertEqual(
            "OpCode DUP is not a valid jump OpCode", str(context.exception)
        )

    def test_emit_call(self):
        sb = vm.ScriptBuilder()
        sb.emit_call(127)
        sb.emit_call(128)

        # captured from C#
        expected = "347f3580000000"
        self.assertEqual(expected, sb.to_array().hex())

    def test_emit_syscall(self):
        sb = vm.ScriptBuilder()
        sb.emit_syscall(0xE393C875)

        # captured from C#
        expected = "4175c893e3"
        self.assertEqual(expected, sb.to_array().hex())

        sb = vm.ScriptBuilder()
        sb.emit_syscall(vm.Syscalls.SYSTEM_CONTRACT_CALL)

        # captured from C#
        expected = "41627d5b52"
        self.assertEqual(expected, sb.to_array().hex())

    def test_emit_contract_call(self):
        """
        UInt160 sh;
        UInt160.TryParse("0xef4073a0f2b305a38ec4050e4d3d28bc40ea63f5", out sh);

        using ScriptBuilder sb = new();
        sb.EmitDynamicCall(sh, "symbol", CallFlags.ReadOnly);
        Console.WriteLine(sb.ToArray().ToHexString());
        """
        sh = types.UInt160.from_string("0xef4073a0f2b305a38ec4050e4d3d28bc40ea63f5")
        sb = vm.ScriptBuilder()
        sb.emit_contract_call(sh, "symbol", callflags.CallFlags.READ_ONLY)

        expected = (
            "c2150c0673796d626f6c0c14f563ea40bc283d4d0e05c48ea305b3f2a07340ef41627d5b52"
        )
        self.assertEqual(expected, sb.to_array().hex())

    def test_emit_contract_call_with_args(self):
        """
        UInt160 sh;
        UInt160 account;
        UInt160.TryParse("0xef4073a0f2b305a38ec4050e4d3d28bc40ea63f5", out sh);
        UInt160.TryParse("0xd2a4cff31913016155e38e474a2c06d08be276cf", out account);

        using ScriptBuilder sb = new();
        sb.EmitDynamicCall(sh, "balanceOf", CallFlags.ReadOnly, account);
        Console.WriteLine(sb.ToArray().ToHexString());
        """
        sh = types.UInt160.from_string("0xef4073a0f2b305a38ec4050e4d3d28bc40ea63f5")
        account = types.UInt160.from_string(
            "0xd2a4cff31913016155e38e474a2c06d08be276cf"
        )
        sb = vm.ScriptBuilder()
        sb.emit_contract_call_with_args(
            sh, "balanceOf", [account], callflags.CallFlags.READ_ONLY
        )

        expected = "0c14cf76e28bd0062c4a478ee35561011319f3cfa4d211c0150c0962616c616e63654f660c14f563ea40bc283d4d0e05c48ea305b3f2a07340ef41627d5b52"
        self.assertEqual(expected, sb.to_array().hex())


class SyscallsTestCase(unittest.TestCase):
    def test_find_by_name(self):
        name = "System.Runtime.BurnGas"
        s = vm.Syscalls.get_by_name(name)
        self.assertEqual(name, s.name)
        self.assertIsNone(vm.Syscalls.get_by_name("fake"))

    def test_find_by_number(self):
        number = 3163314883
        s = vm.Syscalls.get_by_number(number)
        self.assertEqual(number, s.number)
        self.assertIsNone(vm.Syscalls.get_by_number(123))

    def test_all(self):
        self.assertEqual(36, len(list(vm.Syscalls.all())))

    def test_equality(self):
        # allow to compare against ints (should match the syscall number)
        burn_gas_number = 3163314883
        fake_number = 123
        self.assertEqual(burn_gas_number, vm.Syscalls.SYSTEM_RUNTIME_BURN_GAS)
        self.assertNotEqual(fake_number, vm.Syscalls.SYSTEM_RUNTIME_BURN_GAS)

        # allow to compare against strings (should match the syscall name)
        name = "System.Runtime.BurnGas"
        name_wrong = "nope"
        self.assertEqual(name, vm.Syscalls.SYSTEM_RUNTIME_BURN_GAS)
        self.assertNotEqual(name_wrong, vm.Syscalls.SYSTEM_RUNTIME_BURN_GAS)

        # compare against instances
        self.assertEqual(
            vm.Syscalls.SYSTEM_RUNTIME_BURN_GAS, vm.Syscalls.SYSTEM_RUNTIME_BURN_GAS
        )
        self.assertNotEqual(
            vm.Syscalls.SYSTEM_RUNTIME_BURN_GAS, vm.Syscalls.SYSTEM_CONTRACT_CALL
        )

        # compare against byte sequences
        self.assertEqual(
            b"\x56\xe7\xb3\x27", vm.Syscalls.SYSTEM_CRYPTO_CHECK_STANDARD_ACCOUNT
        )
        self.assertNotEqual(
            b"\x01\x01\x01\x01", vm.Syscalls.SYSTEM_CRYPTO_CHECK_STANDARD_ACCOUNT
        )

        self.assertNotEqual(vm.Syscalls.SYSTEM_RUNTIME_BURN_GAS, None)



import unittest
from neo3.network.payloads import verification
from neo3.core import types, cryptography


class WitnessRuleTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        var w = new WitnessRule
        {
            Action = WitnessRuleAction.Allow,
            Condition = new BooleanCondition{ Expression = true }
        };

        Console.WriteLine(((ISerializable)w).Size);
        Console.WriteLine(w.ToArray().ToHexString());
        Console.WriteLine(w.ToJson());
        """
        cls.rule = verification.WitnessRule(
            verification.WitnessRuleAction.ALLOW, verification.ConditionBool(True)
        )

    def test_len(self):
        # captured from C#, see setUpClass() for the capture code
        expected_len = 3
        self.assertEqual(expected_len, len(self.rule))

    def test_serialization(self):
        # captured from C#, see setUpClass() for the capture code
        expected_data = bytes.fromhex("010001")
        self.assertEqual(expected_data, self.rule.to_array())

    def test_deserialization(self):
        # if the serialization() test for this class passes, we can use that as a reference to test deserialization
        # against
        deserialized_rule = verification.WitnessRule.deserialize_from_bytes(
            self.rule.to_array()
        )
        self.assertEqual(self.rule.action, deserialized_rule.action)
        self.assertEqual(self.rule.condition, deserialized_rule.condition)

    def test_to_json(self):
        expected = {
            "action": "Allow",
            "condition": {"type": "Boolean", "expression": True},
        }
        self.assertDictEqual(expected, self.rule.to_json())


class ConditionsTestCase(unittest.TestCase):
    def shortDescription(self):
        # disable docstring printing in test runner
        return None

    def test_and(self):
        """
        var c = new AndCondition
        {
            Expressions = new WitnessCondition[]
            {
                new BooleanCondition { Expression = true },
                new BooleanCondition { Expression = false }
            }
        };
        Console.WriteLine(((ISerializable)c).Size);
        Console.WriteLine(c.ToArray().ToHexString());
        Console.WriteLine(c.ToJson());
        Console.WriteLine(c.Match(null));
        """
        expected_len = 6
        expected_data = bytes.fromhex("020200010000")
        expected_json = {
            "type": "And",
            "expressions": [
                {"type": "Boolean", "expression": True},
                {"type": "Boolean", "expression": False},
            ],
        }

        c = verification.ConditionAnd(
            [verification.ConditionBool(True), verification.ConditionBool(False)]
        )
        self.assertEqual(expected_len, len(c))
        self.assertEqual(expected_data, c.to_array())
        self.assertEqual(expected_json, c.to_json())
        self.assertEqual(c, verification.ConditionAnd.from_json(c.to_json()))

        deserialized_c = verification.ConditionAnd.deserialize_from_bytes(c.to_array())
        self.assertEqual(c.expressions, deserialized_c.expressions)

        with self.assertRaises(ValueError) as context:
            c2 = verification.ConditionAnd([])
            verification.ConditionAnd.deserialize_from_bytes(c2.to_array())
        self.assertEqual("Cannot have 0 expressions", str(context.exception))

    def test_bool(self):
        """
        var c = new BooleanCondition { Expression = true };
        Console.WriteLine(((ISerializable)c).Size);
        Console.WriteLine(c.ToArray().ToHexString());
        Console.WriteLine(c.ToJson());
        """
        expected_len = 2
        expected_data = bytes.fromhex("0001")
        expected_json = {"type": "Boolean", "expression": True}
        c = verification.ConditionBool(True)
        self.assertEqual(expected_len, len(c))
        self.assertEqual(expected_data, c.to_array())
        self.assertEqual(expected_json, c.to_json())
        self.assertEqual(c, verification.ConditionBool.from_json(c.to_json()))

        deserialized_c = verification.ConditionBool.deserialize_from_bytes(c.to_array())
        self.assertEqual(c, deserialized_c)

    def test_not(self):
        """
        var c = new NotCondition { Expression = new BooleanCondition { Expression = true } };
        Console.WriteLine(((ISerializable)c).Size);
        Console.WriteLine(c.ToArray().ToHexString());
        Console.WriteLine(c.ToJson());
        Console.WriteLine(c.Match(null));
        """
        expected_len = 3
        expected_data = bytes.fromhex("010001")
        expected_json = {
            "type": "Not",
            "expression": {"type": "Boolean", "expression": True},
        }
        c = verification.ConditionNot(verification.ConditionBool(True))
        self.assertEqual(expected_len, len(c))
        self.assertEqual(expected_data, c.to_array())
        self.assertEqual(expected_json, c.to_json())
        self.assertEqual(c, verification.ConditionNot.from_json(c.to_json()))

        deserialized_c = verification.ConditionNot.deserialize_from_bytes(c.to_array())
        self.assertEqual(c, deserialized_c)

    def test_or(self):
        """
        var c = new OrCondition
        {
            Expressions = new WitnessCondition[]
            {
                new BooleanCondition { Expression = true },
                new BooleanCondition { Expression = false }
            }
        };
        Console.WriteLine(((ISerializable)c).Size);
        Console.WriteLine(c.ToArray().ToHexString());
        Console.WriteLine(c.ToJson());
        Console.WriteLine(c.Match(null));
        """
        expected_len = 6
        expected_data = bytes.fromhex("030200010000")
        expected_json = {
            "type": "Or",
            "expressions": [
                {"type": "Boolean", "expression": True},
                {"type": "Boolean", "expression": False},
            ],
        }

        c = verification.ConditionOr(
            [verification.ConditionBool(True), verification.ConditionBool(False)]
        )
        self.assertEqual(expected_len, len(c))
        self.assertEqual(expected_data, c.to_array())
        self.assertEqual(expected_json, c.to_json())
        self.assertEqual(c, verification.ConditionOr.from_json(c.to_json()))

        deserialized_c = verification.ConditionOr.deserialize_from_bytes(c.to_array())
        self.assertEqual(c.expressions, deserialized_c.expressions)

        with self.assertRaises(ValueError) as context:
            c2 = verification.ConditionOr([])
            verification.ConditionOr.deserialize_from_bytes(c2.to_array())
        self.assertEqual("Cannot have 0 expressions", str(context.exception))

    def test_by_contract(self):
        """
        var c = new CalledByContractCondition
        {
            Hash = UInt160.Zero
        };
        Console.WriteLine(((ISerializable)c).Size);
        Console.WriteLine(c.ToArray().ToHexString());
        Console.WriteLine(c.ToJson());
        """
        expected_len = 21
        expected_data = bytes.fromhex("280000000000000000000000000000000000000000")
        expected_json = {
            "type": "CalledByContract",
            "hash": "0x0000000000000000000000000000000000000000",
        }

        c = verification.ConditionCalledByContract(types.UInt160.zero())

        self.assertEqual(expected_len, len(c))
        self.assertEqual(expected_data, c.to_array())
        self.assertEqual(expected_json, c.to_json())
        self.assertEqual(
            c, verification.ConditionCalledByContract.from_json(c.to_json())
        )

        deserialized_c = verification.ConditionCalledByContract.deserialize_from_bytes(
            c.to_array()
        )
        self.assertEqual(c.hash_, deserialized_c.hash_)

    def test_by_entry(self):
        """
        var c = new CalledByEntryCondition();
        Console.WriteLine(((ISerializable)c).Size);
        Console.WriteLine(c.ToArray().ToHexString());
        Console.WriteLine(c.ToJson());
        """
        expected_len = 1
        expected_data = bytes.fromhex("20")
        expected_json = {"type": "CalledByEntry"}

        c = verification.ConditionCalledByEntry()

        self.assertEqual(expected_len, len(c))
        self.assertEqual(expected_data, c.to_array())
        self.assertEqual(expected_json, c.to_json())
        self.assertEqual(c, verification.ConditionCalledByEntry.from_json(c.to_json()))

        deserialized_c = verification.ConditionCalledByEntry.deserialize_from_bytes(
            c.to_array()
        )
        self.assertEqual(c, deserialized_c)

    def test_called_by_group(self):
        """
        var c = new CalledByGroupCondition()
        {
            Group = ECPoint.Parse("02158c4a4810fa2a6a12f7d33d835680429e1a68ae61161c5b3fbc98c7f1f17765", ECCurve.Secp256r1)
        };
        Console.WriteLine(((ISerializable)c).Size);
        Console.WriteLine(c.ToArray().ToHexString());
        Console.WriteLine(c.ToJson());
        """
        expected_len = 34
        expected_data = bytes.fromhex(
            "2902158c4a4810fa2a6a12f7d33d835680429e1a68ae61161c5b3fbc98c7f1f17765"
        )
        expected_json = {
            "type": "CalledByGroup",
            "group": "02158c4a4810fa2a6a12f7d33d835680429e1a68ae61161c5b3fbc98c7f1f17765",
        }
        point_data = expected_data[1:]

        group = cryptography.ECPoint.deserialize_from_bytes(point_data)
        c = verification.ConditionCalledByGroup(group)

        self.assertEqual(expected_len, len(c))
        self.assertEqual(expected_data, c.to_array())
        self.assertEqual(expected_json, c.to_json())
        self.assertEqual(c, verification.ConditionCalledByGroup.from_json(c.to_json()))

        deserialized_c = verification.ConditionCalledByGroup.deserialize_from_bytes(
            c.to_array()
        )
        self.assertEqual(c, deserialized_c)

    def test_various(self):
        with self.assertRaises(ValueError) as context:
            verification.WitnessCondition.from_json({"type": "fake_type"})
        self.assertEqual(
            "fake_type cannot be converted to WitnessConditionType",
            str(context.exception),
        )



from __future__ import annotations
from enum import IntEnum
from events import Events  # type: ignore


msgrouter = Events()

# :noindex:


class Size(IntEnum):
    """
    Explicit bytes of memory consumed
    """

    uint8 = 1
    uint16 = 2
    uint32 = 4
    uint64 = 8
    uint160 = 20
    uint256 = 32



from enum import IntFlag


class CallFlags(IntFlag):
    """
    Describes the required call permissions for contract functions.
    """

    NONE = 0
    READ_STATES = 0x1
    WRITE_STATES = 0x02
    ALLOW_CALL = 0x04
    ALLOW_NOTIFY = 0x08
    STATES = READ_STATES | WRITE_STATES
    READ_ONLY = READ_STATES | ALLOW_CALL
    ALL = STATES | ALLOW_CALL | ALLOW_NOTIFY

    @classmethod
    def from_csharp_name(cls, input: str):
        def get(input):
            if input == "None":
                return CallFlags.NONE
            elif input == "ReadStates":
                return CallFlags.READ_STATES
            elif input == "WriteStates":
                return CallFlags.WRITE_STATES
            elif input == "AllowCall":
                return CallFlags.ALLOW_CALL
            elif input == "AllowNotify":
                return CallFlags.ALLOW_NOTIFY
            elif input == "States":
                return CallFlags.STATES
            elif input == "ReadOnly":
                return CallFlags.READ_ONLY
            elif input == "All":
                return CallFlags.ALL
            else:
                raise ValueError(f"{input} is not a valid member of {cls.__name__}")

        flags = [get(flag.strip()) for flag in input.split(",")]

        result = flags[0]
        for flag in flags[1:]:
            result |= flag
        return result



"""
Signing functions for use with `ChainFacade.invoke`.
"""

import os
from dataclasses import dataclass
from neo3.network.payloads import transaction
from typing import Callable, Awaitable
from neo3.wallet import account


@dataclass
class SigningDetails:
    network: int
    #: for which Signer we're adding a witness
    # witness_index: int


# The signing function adds a witness to the provided transaction
# SigningFunction = Callable[[transaction.Transaction, SigningDetails], None]
SigningFunction = Callable[[transaction.Transaction, SigningDetails], Awaitable]


def sign_with_account(acc: account.Account) -> SigningFunction:
    """
    Sign and add a witness using the account and the provided account password.
    """

    async def account_signer(tx: transaction.Transaction, details: SigningDetails):
        # this will automatically add a witness
        acc.sign_tx(tx, details.network)

    return account_signer


def sign_with_ledger() -> SigningFunction:
    raise NotImplementedError


def sign_on_remote_server() -> SigningFunction:
    async def remote_server_signer(
        tx: transaction.Transaction, details: SigningDetails
    ):
        # call some remote API to get the signature
        # and add a witness with the signature
        raise NotImplementedError

    return remote_server_signer


def sign_with_multisig_account(acc: account.Account) -> SigningFunction:
    """
    Sign and add a multi-signature witness.

    This only works for a 1 out of n multi-signature account.

    Args:
        acc: a multi-signature account
    """

    async def account_signer(tx: transaction.Transaction, details: SigningDetails):
        ctx = account.MultiSigContext()
        # this will automatically add a witness
        acc.sign_multisig_tx(tx, ctx, details.network)

    return account_signer


def no_signing() -> SigningFunction:
    """
    Dummy signing function to use with test invocations.
    """

    async def oh_noes(unused: transaction.Transaction, unused2: SigningDetails):
        raise Exception(
            "can't sign with dummy signing function. Did you add a test_signer to ChainFacade?"
        )

    return oh_noes



"""
Classes to interact with the network such as a specialised RPC Client for NEO Node RPC API and a facade for interacting
 with smart contracts over RPC.
"""

from .noderpc import (
    NeoRpcClient,
    JsonRpcError,
    StackItem,
    StackItemType,
)

__all__ = [
    "NeoRpcClient",
    "JsonRpcError",
    "StackItem",
    "StackItemType",
]
